import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
from contextlib import nullcontext as does_not_raise

import tensorflow as tf

from DeepSaki.layers.fourier_layer import FFT2D
from DeepSaki.layers.fourier_layer import FFT3D
from DeepSaki.layers.fourier_layer import FourierConvolution2D
from DeepSaki.layers.fourier_layer import FourierFilter2D
from DeepSaki.layers.fourier_layer import FourierLayer
from DeepSaki.layers.fourier_layer import FourierPooling2D
from DeepSaki.layers.fourier_layer import FrequencyFilter
from DeepSaki.layers.fourier_layer import MultiplicationType
from DeepSaki.layers.fourier_layer import iFFT2D
from DeepSaki.layers.fourier_layer import iFFT3D
from DeepSaki.layers.fourier_layer import rFFT2DFilter
from tests.DeepSaki_test.layers_test.layers_test import CommonLayerChecks
from tests.DeepSaki_test.layers_test.layers_test import DeepSakiLayerChecks


class TestFourierLayer:
    @pytest.fixture()
    def fourier_layer(self):
        return FourierLayer()

    @pytest.mark.parametrize(
        ("input_shape", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3)), tf.TensorShape((1, 3, 16, 16))),
            (tf.TensorShape((1, 8, 16, 16, 3)), tf.TensorShape((1, 3, 8, 16, 16))),
        ],
    )
    def test_change_to_channel_first_correct_shape(self, fourier_layer, input_shape, expected_shape):
        resulting_tensor = fourier_layer._change_to_channel_first(tf.ones(shape=input_shape))
        assert resulting_tensor.shape == expected_shape

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((1, 16, 16, 3)), does_not_raise()),
            (tf.TensorShape((1, 8, 16, 16, 3)), does_not_raise()),
            (tf.TensorShape((1, 8, 16)), pytest.raises(ValueError)),
            (tf.TensorShape((1, 8, 16, 16, 3, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_change_to_channel_first_raises_error(self, fourier_layer, input_shape, expected_context):
        with expected_context:
            _ = fourier_layer._change_to_channel_first(tf.ones(shape=input_shape))

    @pytest.mark.parametrize(
        ("input_shape", "expected_shape"),
        [
            (tf.TensorShape((1, 3, 16, 16)), tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((1, 3, 8, 16, 16)), tf.TensorShape((1, 8, 16, 16, 3))),
        ],
    )
    def test_change_to_channel_last_correct_shape(self, fourier_layer, input_shape, expected_shape):
        resulting_tensor = fourier_layer._change_to_channel_last(tf.ones(shape=input_shape))
        assert resulting_tensor.shape == expected_shape

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((1, 3, 16, 16)), does_not_raise()),
            (tf.TensorShape((1, 3, 8, 16, 16)), does_not_raise()),
            (tf.TensorShape((1, 3, 8)), pytest.raises(ValueError)),
            (tf.TensorShape((1, 3, 8, 16, 16, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_change_to_channel_last_raises_error(self, fourier_layer, input_shape, expected_context):
        with expected_context:
            _ = fourier_layer._change_to_channel_last(tf.ones(shape=input_shape))

    @pytest.mark.parametrize(
        ("batch", "height", "width", "channel", "filter"),
        [
            (1, 16, 16, 3, 5),
            (8, 8, 8, 12, 24),
            (3, 15, 15, 4, 8),
        ],
    )
    def test_matrix_product(
        self,
        fourier_layer,
        batch,
        height,
        width,
        channel,
        filter,
    ):
        image_tensor = tf.ones(shape=(batch, channel, height, width))
        kernel_tensor = tf.ones(shape=(filter, channel, height, width))
        result = fourier_layer._elementwise_product(image_tensor, kernel_tensor)
        expected_tensor = channel * tf.ones(shape=(batch, filter, height, width))
        assert tf.math.reduce_all(result.numpy() == pytest.approx(expected_tensor.numpy(), 0.01))

    @pytest.mark.parametrize(
        ("batch", "height", "width", "channel", "filter"),
        [
            (1, 16, 16, 3, 5),
            (8, 8, 8, 12, 24),
            (3, 15, 15, 4, 8),
        ],
    )
    def test_elementwise_product(
        self,
        fourier_layer,
        batch,
        height,
        width,
        channel,
        filter,
    ):
        image_tensor = tf.ones(shape=(batch, channel, height, width))
        kernel_tensor = tf.ones(shape=(filter, channel, height, width))
        result = fourier_layer._elementwise_product(image_tensor, kernel_tensor)
        expected_tensor = channel * tf.ones(shape=(batch, filter, height, width))
        assert tf.math.reduce_all(result.numpy() == pytest.approx(expected_tensor.numpy(), 0.01))

    @pytest.mark.parametrize(
        ("batch", "height", "width", "channel", "filter"),
        [
            (1, 16, 16, 3, 5),
            (8, 8, 8, 12, 24),
            (3, 15, 15, 4, 8),
        ],
    )
    def test_matrix_product_same_as_element_wise(
        self,
        fourier_layer,
        batch,
        height,
        width,
        channel,
        filter,
    ):
        image_tensor = tf.ones(shape=(batch, channel, height, width))
        kernel_tensor = tf.ones(shape=(filter, channel, height, width))
        result_element_wise = fourier_layer._elementwise_product(image_tensor, kernel_tensor)
        result_matrix_product = fourier_layer._matrix_product(image_tensor, kernel_tensor)
        assert tf.math.reduce_all(result_element_wise.numpy() == pytest.approx(result_matrix_product.numpy(), 0.01))

    @pytest.mark.parametrize(
        ("multiplication_type", "expected_context"),
        [
            (MultiplicationType.ELEMENT_WISE, does_not_raise()),
            (MultiplicationType.MATRIX_PRODUCT, does_not_raise()),
            (None, pytest.raises(AttributeError)),
            ("Any wrong String", pytest.raises(AttributeError)),
            (1, pytest.raises(AttributeError)),
            (2, pytest.raises(AttributeError)),
        ],
    )
    def test_get_multiplication_function_raises_error_wrong_multiplication_type(
        self, fourier_layer, multiplication_type, expected_context
    ):
        with expected_context:
            _ = fourier_layer._get_multiplication_function(multiplication_type=multiplication_type)


class TestFourierConvolution2D(DeepSakiLayerChecks):
    @pytest.fixture()
    def fourier_convolution_2D(self):
        return FourierConvolution2D()

    @pytest.mark.parametrize(
        ("kernels", "expected"),
        [
            ((3, 3), (1, 1)),
            ((5, 5), (2, 2)),
            ((7, 7), (3, 3)),
            ((9, 9), (4, 4)),
            ((5, 3), (2, 1)),
            ((3, 5), (1, 2)),
            ((7, 5), (3, 2)),
            ((5, 7), (2, 3)),
        ],
    )
    def test_get_image_padding(self, fourier_convolution_2D, kernels, expected):
        padding = fourier_convolution_2D._get_image_padding(kernels)
        assert padding == expected

    @pytest.mark.parametrize(
        ("input_shape", "expected_shape"),
        [
            ((1, 2, 8, 8), (1, 2, 8, 8)),
            ((1, 2, 8, 10), (1, 2, 8, 16)),
            ((1, 2, 17, 19), (1, 2, 32, 32)),
            ((8, 5, 50, 64), (8, 5, 64, 64)),
            ((4, 10, 77, 55), (4, 10, 128, 64)),
        ],
    )
    def test_fill_image_shape_power_2(self, fourier_convolution_2D, input_shape, expected_shape):
        output_shape = fourier_convolution_2D._fill_image_shape_power_2(input_shape)
        assert output_shape == expected_shape

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, fourier_convolution_2D, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(fourier_convolution_2D, input_shape, expected_context)

    @pytest.mark.parametrize("kernels", [(3, 3), (9, 9), (16, 16)])
    @pytest.mark.parametrize("use_bias", [True, False])
    @pytest.mark.parametrize("apply_conjugate", [True, False])
    @pytest.mark.parametrize("pad_to_power_2", [True, False])
    @pytest.mark.parametrize("method", [MultiplicationType.ELEMENT_WISE, MultiplicationType.MATRIX_PRODUCT])
    @pytest.mark.parametrize(
        ("input_shape", "filters", "is_channel_first", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3)), 5, False, tf.TensorShape((1, 16, 16, 5))),
            (tf.TensorShape((1, 34, 30, 16)), 9, False, tf.TensorShape((1, 34, 30, 9))),
            (tf.TensorShape((8, 16, 16, 3)), 3, False, tf.TensorShape((8, 16, 16, 3))),
            (tf.TensorShape((1, 5, 16, 16)), 5, True, tf.TensorShape((1, 5, 16, 16))),
            (tf.TensorShape((1, 16, 17, 34)), 9, True, tf.TensorShape((1, 9, 17, 34))),
            (tf.TensorShape((8, 3, 16, 16)), 3, True, tf.TensorShape((8, 3, 16, 16))),
        ],
    )
    def test_call_correct_output_shape(
        self,
        input_shape,
        filters,
        kernels,
        use_bias,
        is_channel_first,
        apply_conjugate,
        pad_to_power_2,
        method,
        expected_shape,
    ):
        layer_instance = FourierConvolution2D(
            filters,
            kernels,
            use_bias,
            is_channel_first,
            apply_conjugate,
            pad_to_power_2,
            method,
        )
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)


class TestFourierFilter2D(DeepSakiLayerChecks):
    @pytest.fixture()
    def fourier_filter_2D(self):
        return FourierFilter2D()

    @pytest.mark.parametrize(
        ("input_shape", "input_complex", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4)), True, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), False, pytest.raises(ValueError)),
            (tf.TensorShape((1, 32, 32, 3)), True, does_not_raise()),
            (tf.TensorShape((8, 64, 64)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), True, pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, fourier_filter_2D, input_shape, input_complex, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(
            fourier_filter_2D, input_shape, expected_context, make_input_complex=input_complex
        )

    @pytest.mark.parametrize("use_bias", [True, False])
    @pytest.mark.parametrize(
        "multiplication_type", [MultiplicationType.ELEMENT_WISE, MultiplicationType.MATRIX_PRODUCT]
    )
    @pytest.mark.parametrize(
        ("input_shape", "filters", "is_channel_first", "expected_shape"),
        [
            (tf.TensorShape((1, 8, 8, 3)), 12, False, tf.TensorShape((1, 8, 8, 12))),
            (tf.TensorShape((8, 3, 8, 8)), 9, True, tf.TensorShape((8, 9, 8, 8))),
        ],
    )
    def test_call_correct_output_shape(
        self, input_shape, filters, is_channel_first, use_bias, multiplication_type, expected_shape
    ):
        layer_instance = FourierFilter2D(filters, use_bias, is_channel_first, multiplication_type)
        CommonLayerChecks.has_call_correct_output_shape(
            layer_instance, input_shape, expected_shape, make_input_complex=True
        )


class TestFFT2D(DeepSakiLayerChecks):
    @pytest.fixture()
    def fft2d(self):
        return FFT2D()

    @pytest.mark.parametrize(
        ("input_shape", "input_complex", "apply_real_fft", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), True, False, does_not_raise()),
            (tf.TensorShape((8, 64, 64, 4)), False, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4)), True, True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4)), False, True, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), True, False, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), False, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64)), True, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), True, False, pytest.raises(ValueError)),
            (tf.TensorShape((8)), True, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), True, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), True, False, pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, input_shape, input_complex, apply_real_fft, expected_context):
        layer_instance = FFT2D(apply_real_fft=apply_real_fft)
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(
            layer_instance, input_shape, expected_context, make_input_complex=input_complex
        )

    @pytest.mark.parametrize("shift_fft", [True, False])
    @pytest.mark.parametrize(
        ("input_shape", "is_channel_first", "apply_real_fft", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 8, 3)), False, False, tf.TensorShape((1, 16, 8, 3))),
            (tf.TensorShape((1, 16, 16, 3)), False, False, tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((1, 34, 34, 16)), False, False, tf.TensorShape((1, 34, 34, 16))),
            (tf.TensorShape((8, 12, 16, 3)), False, False, tf.TensorShape((8, 12, 16, 3))),
            (tf.TensorShape((1, 5, 16, 16)), True, False, tf.TensorShape((1, 5, 16, 16))),
            (tf.TensorShape((1, 16, 34, 34)), True, False, tf.TensorShape((1, 16, 34, 34))),
            (tf.TensorShape((8, 3, 16, 16)), True, False, tf.TensorShape((8, 3, 16, 16))),
            (tf.TensorShape((1, 16, 16, 3)), False, True, tf.TensorShape((1, 16, 9, 3))),
            (tf.TensorShape((1, 34, 18, 16)), False, True, tf.TensorShape((1, 34, 10, 16))),
            (tf.TensorShape((8, 16, 16, 3)), False, True, tf.TensorShape((8, 16, 9, 3))),
            (tf.TensorShape((1, 5, 16, 16)), True, True, tf.TensorShape((1, 5, 16, 9))),
            (tf.TensorShape((1, 16, 34, 34)), True, True, tf.TensorShape((1, 16, 34, 18))),
            (tf.TensorShape((8, 3, 16, 16)), True, True, tf.TensorShape((8, 3, 16, 9))),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, is_channel_first, apply_real_fft, shift_fft, expected_shape):
        layer_instance = FFT2D(is_channel_first, apply_real_fft, shift_fft)
        CommonLayerChecks.has_call_correct_output_shape(
            layer_instance, input_shape, expected_shape, make_input_complex=(not apply_real_fft)
        )


class TestFFT3D(DeepSakiLayerChecks):
    @pytest.fixture()
    def fft3d(self):
        return FFT3D()

    @pytest.mark.parametrize(
        ("input_shape", "input_complex", "apply_real_fft", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), True, False, does_not_raise()),
            (tf.TensorShape((8, 64, 64, 4)), False, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4)), True, True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4)), False, True, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), True, False, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), False, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64)), True, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), True, False, pytest.raises(ValueError)),
            (tf.TensorShape((8)), True, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), True, False, does_not_raise()),
            (tf.TensorShape((8, 64, 64, 4, 5)), False, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), True, True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), False, True, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3, 9)), True, False, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3, 9)), False, False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), True, False, pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, input_shape, input_complex, apply_real_fft, expected_context):
        layer_instance = FFT3D(apply_real_fft=apply_real_fft)
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(
            layer_instance, input_shape, expected_context, make_input_complex=input_complex
        )

    @pytest.mark.parametrize("shift_fft", [True, False])
    @pytest.mark.parametrize(
        ("input_shape", "is_channel_first", "apply_real_fft", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3)), False, False, tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((8, 34, 34, 16)), False, False, tf.TensorShape((8, 34, 34, 16))),
            (tf.TensorShape((1, 5, 16, 16)), True, False, tf.TensorShape((1, 5, 16, 16))),
            (tf.TensorShape((8, 16, 34, 34)), True, False, tf.TensorShape((8, 16, 34, 34))),
            (tf.TensorShape((1, 16, 16, 3)), False, True, tf.TensorShape((1, 16, 9, 3))),
            (tf.TensorShape((8, 34, 34, 16)), False, True, tf.TensorShape((8, 34, 18, 16))),
            (tf.TensorShape((1, 5, 16, 16)), True, True, tf.TensorShape((1, 5, 16, 9))),
            (tf.TensorShape((8, 16, 34, 34)), True, True, tf.TensorShape((8, 16, 34, 18))),
            (tf.TensorShape((1, 8, 16, 16, 3)), False, False, tf.TensorShape((1, 8, 16, 16, 3))),
            (tf.TensorShape((8, 1, 34, 34, 16)), False, False, tf.TensorShape((8, 1, 34, 34, 16))),
            (tf.TensorShape((1, 5, 8, 16, 16)), True, False, tf.TensorShape((1, 5, 8, 16, 16))),
            (tf.TensorShape((8, 16, 1, 34, 34)), True, False, tf.TensorShape((8, 16, 1, 34, 34))),
            (tf.TensorShape((1, 8, 16, 16, 3)), False, True, tf.TensorShape((1, 8, 16, 9, 3))),
            (tf.TensorShape((8, 1, 34, 34, 16)), False, True, tf.TensorShape((8, 1, 34, 18, 16))),
            (tf.TensorShape((1, 5, 8, 16, 16)), True, True, tf.TensorShape((1, 5, 8, 16, 9))),
            (tf.TensorShape((8, 16, 1, 34, 34)), True, True, tf.TensorShape((8, 16, 1, 34, 18))),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, is_channel_first, apply_real_fft, shift_fft, expected_shape):
        layer_instance = FFT3D(is_channel_first, apply_real_fft, shift_fft)
        CommonLayerChecks.has_call_correct_output_shape(
            layer_instance, input_shape, expected_shape, make_input_complex=(not apply_real_fft)
        )


class TestiFFT2D(DeepSakiLayerChecks):
    @pytest.fixture()
    def ifft2d(self):
        return iFFT2D()

    @pytest.mark.parametrize(
        ("input_shape", "input_complex", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4)), True, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), False, pytest.raises(ValueError)),
            (tf.TensorShape((1, 32, 32, 3)), True, does_not_raise()),
            (tf.TensorShape((8, 64, 64)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), True, pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, ifft2d, input_shape, input_complex, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(
            ifft2d, input_shape, expected_context, make_input_complex=input_complex
        )

    @pytest.mark.parametrize("shift_fft", [True, False])
    @pytest.mark.parametrize(
        ("input_shape", "is_channel_first", "apply_real_fft", "fft_length", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3)), False, False, None, tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((1, 34, 34, 16)), False, False, (10, 10), tf.TensorShape((1, 34, 34, 16))),
            (tf.TensorShape((8, 16, 16, 3)), False, False, None, tf.TensorShape((8, 16, 16, 3))),
            (tf.TensorShape((1, 5, 16, 16)), True, False, None, tf.TensorShape((1, 5, 16, 16))),
            (tf.TensorShape((1, 16, 34, 34)), True, False, None, tf.TensorShape((1, 16, 34, 34))),
            (tf.TensorShape((8, 3, 16, 16)), True, False, None, tf.TensorShape((8, 3, 16, 16))),
            (tf.TensorShape((1, 16, 9, 3)), False, True, None, tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((1, 34, 18, 16)), False, True, None, tf.TensorShape((1, 34, 34, 16))),
            (tf.TensorShape((8, 16, 9, 3)), False, True, (32, 32), tf.TensorShape((8, 32, 32, 3))),
            (tf.TensorShape((1, 5, 16, 9)), True, True, None, tf.TensorShape((1, 5, 16, 16))),
            (tf.TensorShape((1, 16, 34, 18)), True, True, None, tf.TensorShape((1, 16, 34, 34))),
            (tf.TensorShape((8, 3, 16, 9)), True, True, (20, 20), tf.TensorShape((8, 3, 20, 20))),
        ],
    )
    def test_call_correct_output_shape(
        self, input_shape, is_channel_first, apply_real_fft, shift_fft, fft_length, expected_shape
    ):
        layer_instance = iFFT2D(is_channel_first, apply_real_fft, shift_fft, fft_length)
        CommonLayerChecks.has_call_correct_output_shape(
            layer_instance, input_shape, expected_shape, make_input_complex=True
        )


class TestiFFT3D(DeepSakiLayerChecks):
    @pytest.fixture()
    def ifft3d(self):
        return iFFT3D()

    @pytest.mark.parametrize(
        ("input_shape", "input_complex", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4)), True, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), False, pytest.raises(ValueError)),
            (tf.TensorShape((1, 32, 32, 3)), True, does_not_raise()),
            (tf.TensorShape((8, 64, 64)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), True, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3, 9)), True, does_not_raise()),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), True, pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, ifft3d, input_shape, input_complex, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(
            ifft3d, input_shape, expected_context, make_input_complex=input_complex
        )

    @pytest.mark.parametrize("shift_fft", [True, False])
    @pytest.mark.parametrize(
        ("input_shape", "is_channel_first", "apply_real_fft", "fft_length", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3)), False, False, None, tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((8, 34, 34, 16)), False, False, (10, 20, 30), tf.TensorShape((8, 34, 34, 16))),
            (tf.TensorShape((1, 5, 16, 16)), True, False, None, tf.TensorShape((1, 5, 16, 16))),
            (tf.TensorShape((8, 16, 34, 34)), True, False, None, tf.TensorShape((8, 16, 34, 34))),
            (tf.TensorShape((1, 16, 9, 3)), False, True, None, tf.TensorShape((1, 16, 16, 3))),
            (tf.TensorShape((8, 34, 18, 16)), False, True, None, tf.TensorShape((8, 34, 34, 16))),
            (tf.TensorShape((1, 5, 16, 9)), True, True, (10, 20, 30), tf.TensorShape((1, 10, 20, 30))),
            (tf.TensorShape((8, 16, 34, 18)), True, True, None, tf.TensorShape((8, 16, 34, 34))),
            (tf.TensorShape((1, 8, 16, 16, 3)), False, False, None, tf.TensorShape((1, 8, 16, 16, 3))),
            (tf.TensorShape((8, 1, 34, 34, 16)), False, False, None, tf.TensorShape((8, 1, 34, 34, 16))),
            (tf.TensorShape((1, 5, 8, 16, 16)), True, False, None, tf.TensorShape((1, 5, 8, 16, 16))),
            (tf.TensorShape((8, 16, 1, 34, 34)), True, False, None, tf.TensorShape((8, 16, 1, 34, 34))),
            (tf.TensorShape((1, 8, 16, 9, 3)), False, True, None, tf.TensorShape((1, 8, 16, 16, 3))),
            (tf.TensorShape((8, 1, 34, 18, 16)), False, True, (10, 20, 30), tf.TensorShape((8, 10, 20, 30, 16))),
            (tf.TensorShape((1, 5, 8, 16, 9)), True, True, None, tf.TensorShape((1, 5, 8, 16, 16))),
            (tf.TensorShape((8, 16, 1, 34, 18)), True, True, None, tf.TensorShape((8, 16, 1, 34, 34))),
        ],
    )
    def test_call_correct_output_shape(
        self, input_shape, is_channel_first, apply_real_fft, shift_fft, fft_length, expected_shape
    ):
        layer_instance = iFFT3D(is_channel_first, apply_real_fft, shift_fft, fft_length)
        CommonLayerChecks.has_call_correct_output_shape(
            layer_instance, input_shape, expected_shape, make_input_complex=True
        )


class TestFourierPooling2D(DeepSakiLayerChecks):
    @pytest.fixture()
    def fourier_pooling_2d(self):
        return FourierPooling2D()

    @pytest.mark.parametrize(
        ("input_shape", "input_complex", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), False, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4)), True, does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), False, pytest.raises(ValueError)),
            (tf.TensorShape((1, 32, 32, 3)), True, does_not_raise()),
            (tf.TensorShape((8, 64, 64)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), True, pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), True, pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, fourier_pooling_2d, input_shape, input_complex, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(
            fourier_pooling_2d, input_shape, expected_context, make_input_complex=input_complex
        )

    @pytest.mark.parametrize(
        ("input_shape", "is_channel_first", "input_from_rfft", "expected_shape"),
        [
            (tf.TensorShape((1, 16, 16, 3)), False, False, tf.TensorShape((1, 8, 8, 3))),
            (tf.TensorShape((8, 34, 34, 9)), False, False, tf.TensorShape((8, 17, 17, 9))),
            (tf.TensorShape((1, 16, 9, 3)), False, True, tf.TensorShape((1, 8, 5, 3))),
            (tf.TensorShape((8, 34, 18, 7)), False, True, tf.TensorShape((8, 17, 9, 7))),
            (tf.TensorShape((1, 3, 16, 16)), True, False, tf.TensorShape((1, 3, 8, 8))),
            (tf.TensorShape((8, 9, 34, 34)), True, False, tf.TensorShape((8, 9, 17, 17))),
            (tf.TensorShape((1, 3, 16, 9)), True, True, tf.TensorShape((1, 3, 8, 5))),
            (tf.TensorShape((8, 7, 34, 18)), True, True, tf.TensorShape((8, 7, 17, 9))),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, is_channel_first, input_from_rfft, expected_shape):
        layer_instance = FourierPooling2D(is_channel_first, input_from_rfft)
        CommonLayerChecks.has_call_correct_output_shape(
            layer_instance, input_shape, expected_shape, make_input_complex=True
        )


class TestrFFT2DFilter(DeepSakiLayerChecks):
    @pytest.fixture()
    def rfft2d_filter(self):
        return rFFT2DFilter()

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, rfft2d_filter, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(rfft2d_filter, input_shape, expected_context)

    @pytest.mark.parametrize("filter_type", [FrequencyFilter.LOW_PASS, FrequencyFilter.HIGH_PASS])
    @pytest.mark.parametrize(
        ("input_shape", "is_channel_first"),
        [
            (tf.TensorShape((1, 16, 16, 3)), False),
            (tf.TensorShape((1, 3, 30, 30)), True),
            (tf.TensorShape((8, 32, 32, 9)), False),
            (tf.TensorShape((8, 9, 25, 25)), True),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, is_channel_first, filter_type):
        layer_instance = rFFT2DFilter(is_channel_first, filter_type)
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, input_shape)



import os
from contextlib import nullcontext as does_not_raise

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.models.autoencoders import ResNet
from DeepSaki.models.autoencoders import UNet
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_bottleneck
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_decoder
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_encoder
from tests.DeepSaki_test.models_test.models_test import CommonModelChecks
from tests.DeepSaki_test.models_test.models_test import DeepSakiModelChecks
from DeepSaki.types.layers_enums import LinearLayerType


@pytest.mark.parametrize(("autoencoder_model"), [UNet, ResNet])
class TestGenericAutoEncoder(DeepSakiModelChecks):
    @pytest.fixture(autouse=True)
    def mock_sub_models(self, mocker):
        calling_module = "DeepSaki.models.autoencoders"
        _mock_encoder(mocker, calling_module)
        _mock_bottleneck(mocker, calling_module)
        _mock_decoder(mocker, calling_module)

    @pytest.fixture()
    def model_instance(self, autoencoder_model):
        return autoencoder_model(number_of_levels=2, filters=8, number_of_blocks=1)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, model_instance, input_shape, expected_context):
        CommonModelChecks.does_call_raises_error_wrong_input_spec(model_instance, input_shape, expected_context)

    @pytest.mark.parametrize("linear_layer_type", [LinearLayerType.MLP, LinearLayerType.CONV_1x1])
    @pytest.mark.parametrize("number_of_levels", [2, 3, 4])
    @pytest.mark.parametrize("filters", [8, 16])
    @pytest.mark.parametrize(
        ("input_shape"),
        [
            tf.TensorShape((1, 64, 64, 3)),
            tf.TensorShape((8, 64, 64, 4)),
        ],
    )
    def test_call_correct_output_shape(
        self, autoencoder_model, input_shape, number_of_levels, filters, linear_layer_type
    ):
        model_instance = autoencoder_model(
            number_of_levels=number_of_levels, filters=filters, linear_layer_type=linear_layer_type
        )
        CommonModelChecks.has_call_correct_output_shape(model_instance, input_shape, input_shape)

    @pytest.mark.parametrize(
        ("linear_layer_type", "expected_context"),
        [
            (LinearLayerType.MLP, does_not_raise()),
            (LinearLayerType.CONV_1x1, does_not_raise()),
            ("Any other String", pytest.raises(ValueError)),
        ],
    )
    def test_init_raises_wrong_linear_layer_type_type(self, autoencoder_model, linear_layer_type, expected_context):
        with expected_context:
            _ = autoencoder_model(number_of_levels=2, filters=8, linear_layer_type=linear_layer_type)



import os
from contextlib import nullcontext as does_not_raise

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.models.discriminators import LayoutContentDiscriminator
from DeepSaki.models.discriminators import PatchDiscriminator
from DeepSaki.models.discriminators import UNetDiscriminator
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_bottleneck
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_decoder
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_encoder
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_global_sum_pooling_2d
from tests.DeepSaki_test.layers_test.mocked_layers import _mock_scalar_gated_self_attention
from tests.DeepSaki_test.models_test.models_test import CommonModelChecks
from tests.DeepSaki_test.models_test.models_test import DeepSakiModelChecks
from DeepSaki.types.layers_enums import LinearLayerType


class TestLayoutContentDiscriminator(DeepSakiModelChecks):
    @pytest.fixture()
    def model_instance(self):
        return LayoutContentDiscriminator(filters=64, number_of_blocks=2)

    @pytest.fixture(autouse=True)
    def mock_sub_models(self, mocker):
        calling_module = "DeepSaki.models.discriminators"
        _mock_encoder(mocker, calling_module)
        _mock_scalar_gated_self_attention(mocker, calling_module)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((1, 256, 256, 4)), does_not_raise()),
            (tf.TensorShape((1, 256, 256, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, model_instance, input_shape, expected_context):
        CommonModelChecks.does_call_raises_error_wrong_input_spec(model_instance, input_shape, expected_context)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 256, 256, 4)), does_not_raise()),
            (tf.TensorShape((1, 344, 344, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64, 3)), pytest.raises(ValueError)),
            (tf.TensorShape((1, 256, 64, 3)), pytest.raises(ValueError)),
            (tf.TensorShape((1, 64, 256, 3)), pytest.raises(ValueError)),
        ],
    )
    def test_build_raises_error_wrong_input_height_width(self, model_instance, input_shape, expected_context):
        with expected_context:
            model_instance.build(input_shape)

    @pytest.mark.parametrize("linear_layer_type", [LinearLayerType.MLP, LinearLayerType.CONV_1x1])
    @pytest.mark.parametrize("use_self_attention", [True, False])
    @pytest.mark.parametrize("filters", [8, 16])
    @pytest.mark.parametrize(
        ("input_shape"),
        [
            tf.TensorShape((1, 512, 512, 3)),
            tf.TensorShape((8, 256, 256, 4)),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, use_self_attention, filters, linear_layer_type):
        model_instance = LayoutContentDiscriminator(
            use_self_attention=use_self_attention, filters=filters, linear_layer_type=linear_layer_type
        )
        content_output, layout_output = model_instance(tf.ones(shape=input_shape))
        expected_shape_layout_output = [
            input_shape[0],
            input_shape[1] // 2**3,  # 3 is the number of downsampling blocks of the encoder
            input_shape[2] // 2**3,
            1,
        ]
        expected_shape_content_output = [
            input_shape[0],
            1,
            1,
            filters * 2**3,  # 3 is the number of downsampling blocks of the encoder
        ]

        assert layout_output.shape == expected_shape_layout_output
        assert content_output.shape == expected_shape_content_output


class TestPatchDiscriminator(DeepSakiModelChecks):
    @pytest.fixture()
    def model_instance(self):
        return PatchDiscriminator(filters=8, number_of_blocks=1)

    @pytest.fixture(autouse=True)
    def mock_sub_models(self, mocker):
        calling_module = "DeepSaki.models.discriminators"
        _mock_encoder(mocker, calling_module)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, model_instance, input_shape, expected_context):
        CommonModelChecks.does_call_raises_error_wrong_input_spec(model_instance, input_shape, expected_context)

    @pytest.mark.parametrize("num_down_blocks", [2, 3, 4])
    @pytest.mark.parametrize("filters", [8, 16])
    @pytest.mark.parametrize(
        ("input_shape"),
        [
            tf.TensorShape((1, 64, 64, 3)),
            tf.TensorShape((8, 64, 64, 4)),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, num_down_blocks, filters):
        model_instance = PatchDiscriminator(num_down_blocks=num_down_blocks, filters=filters)
        expected_shape = [
            input_shape[0],
            input_shape[1] // 2**num_down_blocks,
            input_shape[2] // 2**num_down_blocks,
            1,
        ]
        CommonModelChecks.has_call_correct_output_shape(model_instance, input_shape, expected_shape)


class TestUNetDiscriminator(DeepSakiModelChecks):
    @pytest.fixture()
    def model_instance(self):
        return UNetDiscriminator(number_of_levels=2, filters=8, number_of_blocks=1)

    @pytest.fixture(autouse=True)
    def mock_sub_models(self, mocker):
        calling_module = "DeepSaki.models.discriminators"
        _mock_encoder(mocker, calling_module)
        _mock_bottleneck(mocker, calling_module)
        _mock_decoder(mocker, calling_module)
        _mock_global_sum_pooling_2d(mocker, calling_module)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, model_instance, input_shape, expected_context):
        CommonModelChecks.does_call_raises_error_wrong_input_spec(model_instance, input_shape, expected_context)

    @pytest.mark.parametrize("linear_layer_type", [LinearLayerType.MLP, LinearLayerType.CONV_1x1])
    @pytest.mark.parametrize("number_of_levels", [2, 3, 4])
    @pytest.mark.parametrize("filters", [8, 16])
    @pytest.mark.parametrize(
        ("input_shape"),
        [
            tf.TensorShape((1, 64, 64, 3)),
            tf.TensorShape((8, 64, 64, 4)),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, number_of_levels, filters, linear_layer_type):
        model_instance = UNetDiscriminator(
            number_of_levels=number_of_levels, filters=filters, linear_layer_type=linear_layer_type
        )
        global_output, decoder_output = model_instance(tf.ones(shape=input_shape))
        expected_shape_decoder_output = [*input_shape[0:-1], 1]
        expected_shape_global_output = [input_shape[0], 1]

        assert decoder_output.shape == expected_shape_decoder_output
        assert global_output.shape == expected_shape_global_output



from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

import tensorflow as tf
import tensorflow_addons as tfa

from DeepSaki.constraints import NonNegative
from DeepSaki.layers.layer_helper import dropout_func
from DeepSaki.layers.layer_helper import pad_func
from DeepSaki.types.layers_enums import DownSampleType
from DeepSaki.types.layers_enums import PaddingType
from DeepSaki.types.layers_enums import UpSampleType

class Conv2DSplitted(tf.keras.layers.Layer):
    """Convolution layer where a single convolution is splitted into two consecutive convolutions.

    To decrease the number of parameters, a convolution with the kernel_size (kernel,kernel) can be splitted into two
    consecutive convolutions with the kernel_size (kernel,1) and (1,kernel) respectivly.
    """

    def __init__(
        self,
        filters: int = 3,
        kernels: int = 3,
        use_spec_norm: bool = False,
        strides: Tuple[int, int] = (1, 1),
        use_bias: bool = True,
        padding: str = "valid",
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initialize the `Conv2DSplitted` object.

        Args:
            filters (int, optional): Number of filters in the output feature map. Defaults to 3.
            kernels (int, optional): Size of the convolutions' kernels, which will be translated to (kernels, 1) and
                (1,kernels) for the first and seccond convolution respectivly. Defaults to 3.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults
                to False.
            strides (Tuple[int, int], optional): Strides of the convolution layers. Defaults to (1, 1).
            use_bias (bool, optional): Whether or not to use bias weights. Defaults to True.
            padding (str, optional): ["valid"|"same"]
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions
                kernels. Defaults to None.
        """
        super(Conv2DSplitted, self).__init__()
        self.filters = filters
        self.kernels = kernels
        self.use_spec_norm = use_spec_norm
        self.strides = strides
        self.use_bias = use_bias
        self.padding = padding
        self.kernel_initializer = kernel_initializer
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        self.conv1 = tf.keras.layers.Conv2D(
            filters=filters,
            kernel_size=(kernels, 1),
            kernel_initializer=kernel_initializer,
            use_bias=use_bias,
            padding=padding,
            strides=(strides[0], 1),
        )
        self.conv2 = tf.keras.layers.Conv2D(
            filters=filters,
            kernel_size=(1, kernels),
            kernel_initializer=kernel_initializer,
            use_bias=use_bias,
            padding=padding,
            strides=(1, strides[1]),
        )

        if use_spec_norm:
            self.conv1 = tfa.layers.SpectralNormalization(self.conv1)
            self.conv2 = tfa.layers.SpectralNormalization(self.conv2)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `Conv2DSplitted` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape `(batch, height, width, channel)`.

        Returns:
            Convoluted tensor of shape `(batch, H, W, filters)`, where `H` and `W` depend on the padding type used in
                the convolution layers.
        """
        x = self.conv1(inputs)
        return self.conv2(x)

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(Conv2DSplitted, self).get_config()
        config.update(
            {
                "filters": self.filters,
                "kernels": self.kernels,
                "use_spec_norm": self.use_spec_norm,
                "strides": self.strides,
                "use_bias": self.use_bias,
                "padding": self.padding,
                "kernel_initializer": self.kernel_initializer,
            }
        )
        return config


# TODO: Normalization should be setable
# TODO: how to deal with dropout func?
class Conv2DBlock(tf.keras.layers.Layer):
    """Wraps a two-dimensional convolution into a more complex building block."""

    def __init__(
        self,
        filters: int = 3,
        kernels: int = 3,
        split_kernels: bool = False,
        number_of_blocks: int = 1,
        activation: str = "leaky_relu",
        dropout_rate: float = 0.0,
        final_activation: bool = True,
        use_spec_norm: bool = False,
        strides: Tuple[int, int] = (1, 1),
        padding: PaddingType = PaddingType.ZERO,
        apply_final_normalization: bool = True,
        use_bias: bool = True,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `Conv2DBlock` layer.

        Args:
            filters (int, optional): Number of individual filters. Defaults to 3.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)`
                and `(1,kernel)` respectivly. Defaults to False.
            number_of_blocks (int, optional): Number of consecutive convolutional building blocks, i.e. `Conv2DBlock`.
                Defaults to 1.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.0.
            final_activation (bool, optional): Whether or not to activate the output of this layer. Defaults to True.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers.
                Defaults to False.
            strides (Tuple[int, int], optional): Stride of the filter. Defaults to (1, 1).
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            apply_final_normalization (bool, optional): Whether or not to place a normalization on the layer's output.
                Defaults to True.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(Conv2DBlock, self).__init__()
        self.filters = filters
        self.kernels = kernels
        self.split_kernels = split_kernels
        self.number_of_blocks = number_of_blocks
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.final_activation = final_activation
        self.use_spec_norm = use_spec_norm
        self.strides = strides
        self.padding = padding
        self.apply_final_normalization = apply_final_normalization
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        pad = (kernels - 1) // 2  # assumes odd kernel size, which is typical!

        self.blocks = []
        for block in range(number_of_blocks):
            layers = []
            if pad != 0 and self.padding != PaddingType.NONE:
                layers.append(pad_func(pad_values=(pad, pad), padding_type=self.padding))
            layers.append(self._get_conv_layer())

            if block != (self.number_of_blocks - 1) or self.apply_final_normalization:
                layers.append(tfa.layers.InstanceNormalization(gamma_initializer=gamma_initializer))  #

            if block != (self.number_of_blocks - 1) or self.final_activation:
                layers.append(tf.keras.layers.Activation(self.activation))

            self.blocks.append(layers)

        self.dropout = dropout_func(filters, dropout_rate)

    def _get_conv_layer(self) -> tf.keras.layers.Layer:
        """Helper to return correct Conv layer based on `Conv2DBlock` class properties."""
        if self.split_kernels:
            return Conv2DSplitted(
                filters=self.filters,
                kernels=self.kernels,
                use_bias=self.use_bias,
                strides=self.strides,
                use_spec_norm=self.use_spec_norm,
            )
        if self.use_spec_norm:
            return tfa.layers.SpectralNormalization(
                tf.keras.layers.Conv2D(
                    filters=self.filters,
                    kernel_size=(self.kernels, self.kernels),
                    kernel_initializer=self.kernel_initializer,
                    use_bias=self.use_bias,
                    strides=self.strides,
                )
            )
        return tf.keras.layers.Conv2D(
            filters=self.filters,
            kernel_size=(self.kernels, self.kernels),
            kernel_initializer=self.kernel_initializer,
            use_bias=self.use_bias,
            strides=self.strides,
        )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `Conv2DBlock` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (b, h, w, c)

        Returns:
            Tensor of shape (b, h/(`strides[0]`**`number_of_blocks`), w/(`strides[1]`**`number_of_blocks`), `filters`).
        """
        x = inputs

        for block in self.blocks:
            for layer in block:
                x = layer(x)

        if self.dropout_rate > 0:
            x = self.dropout(x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(Conv2DBlock, self).get_config()
        config.update(
            {
                "filters": self.filters,
                "kernels": self.kernels,
                "split_kernels": self.split_kernels,
                "number_of_blocks": self.number_of_blocks,
                "activation": self.activation,
                "dropout_rate": self.dropout_rate,
                "final_activation": self.final_activation,
                "use_spec_norm": self.use_spec_norm,
                "strides": self.strides,
                "padding": self.padding,
                "apply_final_normalization": self.apply_final_normalization,
                "use_bias": self.use_bias,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class DenseBlock(tf.keras.layers.Layer):
    """Wraps a dense layer into a more complex building block."""

    def __init__(
        self,
        units: int,
        number_of_blocks: int = 1,
        activation: str = "leaky_relu",
        dropout_rate: float = 0.0,
        final_activation: bool = True,
        use_spec_norm: bool = False,
        apply_final_normalization: bool = True,
        use_bias: bool = True,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `DenseBlock` layer.

        Args:
            units (int): Number of units of each dense block
            number_of_blocks (int, optional): Number of consecutive subblocks. Defaults to 1.
            activation (str, optional): String literal to obtain activation function. Defaults to "leaky_relu".
            dropout_rate (float, optional): Probability of the dropout layer dropping weights. If the preceeding layer
                has more than one channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.0.
            final_activation (bool, optional): Whether or not to activate the output of this layer. Defaults to True.
            use_spec_norm (bool, optional): Applies spectral normalization to dense layers. Defaults to False.
            apply_final_normalization (bool, optional): Whether or not to apply normalization at the layer's output.
                Defaults to True.
            use_bias (bool, optional): Whether dense layers include bias weights. Defaults to True.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(DenseBlock, self).__init__()
        self.units = units
        self.number_of_blocks = number_of_blocks
        self.dropout_rate = dropout_rate
        self.activation = activation
        self.final_activation = final_activation
        self.use_spec_norm = use_spec_norm
        self.apply_final_normalization = apply_final_normalization
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.input_spec = tf.keras.layers.InputSpec(min_ndim=2)

        self.blocks = []
        for block in range(number_of_blocks):
            layers = []
            layers.append(self._get_dense_layer())

            if block != (number_of_blocks - 1) or self.apply_final_normalization:
                layers.append(tfa.layers.InstanceNormalization(gamma_initializer=gamma_initializer))

            if block != (number_of_blocks - 1) or self.final_activation:
                layers.append(tf.keras.layers.Activation(self.activation))

            self.blocks.append(layers)

        self.dropout = tf.keras.layers.Dropout(dropout_rate)

    def _get_dense_layer(self) -> tf.keras.layers.Layer:
        """Helper to return correct Conv layer based on `DenseBlock` class properties."""
        if self.use_spec_norm:
            return tfa.layers.SpectralNormalization(
                tf.keras.layers.Dense(
                    units=self.units, use_bias=self.use_bias, kernel_initializer=self.kernel_initializer
                )
            )
        return tf.keras.layers.Dense(
            units=self.units, use_bias=self.use_bias, kernel_initializer=self.kernel_initializer
        )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `DenseBlock` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (b, h, w, c)

        Returns:
            Tensor of shape (b, h, w, `units`).
        """
        x = inputs

        for block in self.blocks:
            for layer in block:
                x = layer(x)

        if self.dropout_rate > 0:
            x = self.dropout(x)
        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(DenseBlock, self).get_config()
        config.update(
            {
                "units": self.units,
                "number_of_blocks": self.number_of_blocks,
                "dropout_rate": self.dropout_rate,
                "activation": self.activation,
                "final_activation": self.final_activation,
                "use_spec_norm": self.use_spec_norm,
                "apply_final_normalization": self.apply_final_normalization,
                "use_bias": self.use_bias,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class DownSampleBlock(tf.keras.layers.Layer):
    """Spatial down-sampling for grid-like data using `DeepSaki.layers.ConvBlock2D()`."""

    def __init__(
        self,
        downsampling: DownSampleType = DownSampleType.AVG_POOLING,
        activation: str = "leaky_relu",
        kernels: int = 3,
        use_spec_norm: bool = False,
        padding: PaddingType = PaddingType.ZERO,
        use_bias: bool = True,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `DownSampleBlock` layer.

        Args:
            downsampling (DownSampleType, optional): Downsampling method used. Defaults to DownSampleType.AVG_POOLING.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults
                to False.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(DownSampleBlock, self).__init__()
        self.kernels = kernels
        self.downsampling = downsampling
        self.activation = activation
        self.use_spec_norm = use_spec_norm
        self.padding = padding
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

    def _space_to_depth_block_size_2(self, tensor: tf.Tensor) -> tf.Tensor:
        return tf.nn.space_to_depth(tensor, block_size=2)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.

        Raises:
            ValueError: if unsupported downsampling is provided.
        """
        super(DownSampleBlock, self).build(input_shape)

        self.layers = []
        if self.downsampling == DownSampleType.CONV_STRIDE_2:
            self.layers.append(
                Conv2DBlock(
                    input_shape[-1],
                    self.kernels,
                    activation=self.activation,
                    strides=(2, 2),
                    use_spec_norm=self.use_spec_norm,
                    padding=self.padding,
                    use_bias=self.use_bias,
                    kernel_initializer=self.kernel_initializer,
                    gamma_initializer=self.gamma_initializer,
                )
            )
        elif self.downsampling == DownSampleType.MAX_POOLING:
            # Only spatial downsampling, increase in features is done by the conv2D_block specified later!
            self.layers.append(tf.keras.layers.MaxPooling2D(pool_size=(2, 2)))
        elif self.downsampling == DownSampleType.AVG_POOLING:
            self.layers.append(tf.keras.layers.AveragePooling2D(pool_size=(2, 2)))
        elif self.downsampling == DownSampleType.SPACE_TO_DEPTH:
            self.layers.append(self._space_to_depth_block_size_2)
            self.layers.append(
                tf.keras.layers.Conv2D(
                    filters=input_shape[-1],
                    kernel_size=1,
                    activation=self.activation,
                    use_bias=False,
                    padding="same",
                    kernel_initializer=self.kernel_initializer,
                )
            )
        else:
            raise ValueError("Undefined downsampling provided")

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `DownSampleBlock` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (b,h,w,c)

        Returns:
            Spatially downsampled tensor of shape (b,h/2,w/2,c)
        """
        x = inputs
        for layer in self.layers:
            x = layer(x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(DownSampleBlock, self).get_config()
        config.update(
            {
                "kernels": self.kernels,
                "downsampling": self.downsampling,
                "activation": self.activation,
                "use_spec_norm": self.use_spec_norm,
                "padding": self.padding,
                "use_bias": self.use_bias,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class UpSampleBlock(tf.keras.layers.Layer):
    """Spatial up-sampling for grid-like data using `DeepSaki.layers.ConvBlock2D()`."""

    def __init__(
        self,
        upsampling: UpSampleType = UpSampleType.RESAMPLE_AND_CONV,
        activation: str = "leaky_relu",
        kernels: int = 3,
        use_spec_norm: bool = False,
        use_bias: bool = True,
        padding: PaddingType = PaddingType.ZERO,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `UpSampleBlock` layer.

        Args:
            upsampling (UpSampleType, optional): _description_. Defaults to UpSampleType.RESAMPLE_AND_CONV.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults
                to False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(UpSampleBlock, self).__init__()
        self.kernels = kernels
        self.activation = activation
        self.use_spec_norm = use_spec_norm
        self.upsampling = upsampling
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.padding = padding
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

    def _depth_to_space_block_size_2(self, tensor: tf.Tensor) -> tf.Tensor:
        return tf.nn.depth_to_space(tensor, block_size=2)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.

        Raises:
            ValueError: if `upsampling` is not supported.
        """
        super(UpSampleBlock, self).build(input_shape)
        self.layers = []

        if self.upsampling == UpSampleType.RESAMPLE_AND_CONV:
            self.layers.append(tf.keras.layers.UpSampling2D(interpolation="bilinear"))
            self.layers.append(
                Conv2DBlock(
                    filters=input_shape[-1],
                    kernels=1,
                    number_of_blocks=1,
                    activation=self.activation,
                    use_spec_norm=self.use_spec_norm,
                    use_bias=self.use_bias,
                    padding=self.padding,
                    kernel_initializer=self.kernel_initializer,
                    gamma_initializer=self.gamma_initializer,
                )
            )
        elif self.upsampling == UpSampleType.TRANSPOSE_CONV:
            self.layers.append(
                tf.keras.layers.Conv2DTranspose(
                    input_shape[-1],
                    kernel_size=(self.kernels, self.kernels),
                    strides=(2, 2),
                    kernel_initializer=self.kernel_initializer,
                    padding="same",
                    use_bias=self.use_bias,
                )
            )
            self.layers.append(tfa.layers.InstanceNormalization(gamma_initializer=self.gamma_initializer))
            self.layers.append(tf.keras.layers.Activation(self.activation))
        elif self.upsampling == UpSampleType.DEPTH_TO_SPACE:
            self.layers.append(
                tf.keras.layers.Conv2D(
                    filters=4 * input_shape[-1],
                    kernel_size=1,
                    activation=self.activation,
                    use_bias=False,
                    padding="same",
                    kernel_initializer=self.kernel_initializer,
                )
            )
            self.layers.append(self._depth_to_space_block_size_2)
        else:
            raise ValueError("Undefined upsampling provided")

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `UpSampleBlock` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (b,h,w,c)

        Returns:
            Tensor of shape (b,h*2,w*2,c)
        """
        x = inputs
        for layer in self.layers:
            x = layer(x)
        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(UpSampleBlock, self).get_config()
        config.update(
            {
                "kernels": self.kernels,
                "activation": self.activation,
                "use_spec_norm": self.use_spec_norm,
                "upsampling": self.upsampling,
                "use_bias": self.use_bias,
                "padding": self.padding,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class ResidualBlock(tf.keras.layers.Layer):
    """Residual block with configurable cardinality."""

    def __init__(
        self,
        filters: int = 3,
        kernels: int = 3,
        activation: str = "leaky_relu",
        number_of_blocks: int = 1,
        use_spec_norm: bool = False,
        residual_cardinality: int = 1,
        dropout_rate: float = 0.0,
        use_bias: bool = True,
        padding: PaddingType = PaddingType.ZERO,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `ResidualBlock` layer.

        Args:
            filters (int, optional): Number of individual filters. Defaults to 3.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            number_of_blocks (int, optional): Number of residual subblocks. Defaults to 1.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults
                to False.
            residual_cardinality (int, optional): Number of parallel paths. Defaults to 1.
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.0.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(ResidualBlock, self).__init__()
        self.activation = activation
        self.filters = filters
        self.kernels = kernels
        self.number_of_blocks = number_of_blocks
        self.use_spec_norm = use_spec_norm
        self.residual_cardinality = residual_cardinality
        self.dropout_rate = dropout_rate
        self.use_bias = use_bias
        self.padding = padding
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

        self.pad = int((kernels - 1) / 2)  # assumes odd kernel size, which is typical!

        if residual_cardinality > 1:
            self.intermediateFilters = int(max(filters / 32, filters / 16, filters / 8, filters / 4, filters / 2, 1))
        else:
            self.intermediateFilters = int(max(filters / 4, filters / 2, 1))

        # for each block, add several con
        self.blocks = []
        for _ in range(number_of_blocks):
            cardinals = [
                [
                    Conv2DBlock(
                        filters=self.intermediateFilters,
                        kernels=1,
                        split_kernels=False,
                        number_of_blocks=1,
                        activation=activation,
                        use_spec_norm=use_spec_norm,
                        use_bias=use_bias,
                        padding=padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    ),
                    Conv2DBlock(
                        filters=self.intermediateFilters,
                        kernels=kernels,
                        split_kernels=False,
                        number_of_blocks=1,
                        activation=activation,
                        padding=PaddingType.NONE,
                        use_spec_norm=use_spec_norm,
                        use_bias=use_bias,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    ),
                    Conv2DBlock(
                        filters=filters,
                        kernels=1,
                        split_kernels=False,
                        number_of_blocks=1,
                        activation=activation,
                        use_spec_norm=use_spec_norm,
                        use_bias=use_bias,
                        padding=padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    ),
                ]
                for _ in range(residual_cardinality)
            ]
            self.blocks.append(cardinals)
        self.dropout = dropout_func(filters, dropout_rate)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(ResidualBlock, self).build(input_shape)
        self.conv0 = None
        if input_shape[-1] != self.filters:
            self.conv0 = Conv2DBlock(
                filters=self.filters,
                kernels=1,
                split_kernels=False,
                number_of_blocks=1,
                activation=self.activation,
                use_spec_norm=self.use_spec_norm,
                use_bias=self.use_bias,
                padding=self.padding,
                kernel_initializer=self.kernel_initializer,
                gamma_initializer=self.gamma_initializer,
            )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `ResidualBlock` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (batch, height, width, channel)

        Returns:
            Tensor of shape (batch, height, width, `filters`)
        """
        x = inputs

        if self.conv0 is not None:
            x = self.conv0(x)

        for block in range(self.number_of_blocks):
            residual = x

            if self.pad != 0 and self.padding != PaddingType.NONE:
                x = pad_func(pad_values=(self.pad, self.pad), padding_type=self.padding)(x)

            # y is allways the footpoint of the cardinality blocks
            y = x
            for cardinal in range(self.residual_cardinality):
                # after the first iteration c is the previous cardinality block output
                # its used to iterativly add the result, rather than summing all at once.
                # performance reasons, since otherwise multiple arrays must be stored at once!
                c = x
                x = self.blocks[block][cardinal][0](y)
                x = self.blocks[block][cardinal][1](x)
                x = self.blocks[block][cardinal][2](x)
                if cardinal > 0:
                    x = tf.keras.layers.Add()([x, c])

            x = tf.keras.layers.Add()([x, residual])

        if self.dropout_rate > 0:
            x = self.dropout(x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(ResidualBlock, self).get_config()
        config.update(
            {
                "activation": self.activation,
                "filters": self.filters,
                "kernels": self.kernels,
                "number_of_blocks": self.number_of_blocks,
                "use_spec_norm": self.use_spec_norm,
                "residual_cardinality": self.residual_cardinality,
                "dropout_rate": self.dropout_rate,
                "use_bias": self.use_bias,
                "padding": self.padding,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class ResBlockDown(tf.keras.layers.Layer):
    """Spatial down-sampling with residual connection for grid-like data using `DeepSaki.layers.ConvBlock2D()`.

    Architecture:
        ```mermaid
        flowchart LR
            i([input_tensor])--> c1 & c2
            subgraph ResBlockDown
                subgraph Block[2x]
                c1[ds.layers.Conv2DBlock]
                end
                c1-->ap1[AveragePooling2D]
                c2[ds.layers.Conv2DBlock]-->ap2[AveragePooling2D]
                ap1 & ap2-->add((+))
            end
            add-->o([output_tensor])
        ```

    """

    def __init__(
        self,
        activation: str = "leaky_relu",
        use_spec_norm: bool = False,
        use_bias: bool = True,
        padding: PaddingType = PaddingType.ZERO,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `ResBlockDown` layer.

        Args:
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults
                to False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(ResBlockDown, self).__init__()
        self.activation = activation
        self.use_spec_norm = use_spec_norm
        self.use_bias = use_bias
        self.padding = padding
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(ResBlockDown, self).build(input_shape)

        self.convRes = Conv2DBlock(
            filters=input_shape[-1],
            kernels=1,
            split_kernels=False,
            number_of_blocks=1,
            activation=self.activation,
            use_spec_norm=self.use_spec_norm,
            use_bias=self.use_bias,
            padding=self.padding,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )
        self.conv1 = Conv2DBlock(
            filters=input_shape[-1],
            kernels=3,
            split_kernels=False,
            number_of_blocks=1,
            activation=self.activation,
            use_spec_norm=self.use_spec_norm,
            use_bias=self.use_bias,
            padding=self.padding,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )
        self.conv2 = Conv2DBlock(
            filters=input_shape[-1],
            kernels=3,
            split_kernels=False,
            number_of_blocks=1,
            activation=self.activation,
            use_spec_norm=self.use_spec_norm,
            use_bias=self.use_bias,
            padding=self.padding,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `ResBlockDown` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (b,h,w,c)

        Returns:
            Spatially downsampled tensor of shape (b,h/2,w/2,c)
        """
        path1 = inputs
        path2 = inputs

        path1 = self.convRes(path1)
        path1 = tf.keras.layers.AveragePooling2D()(path1)

        path2 = self.conv1(path2)
        path2 = self.conv2(path2)
        path2 = tf.keras.layers.AveragePooling2D()(path2)

        return tf.keras.layers.Add()([path1, path2])

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(ResBlockDown, self).get_config()
        config.update(
            {
                "activation": self.activation,
                "use_spec_norm": self.use_spec_norm,
                "use_bias": self.use_bias,
                "padding": self.padding,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class ResBlockUp(tf.keras.layers.Layer):
    """Spatial up-sampling with residual connection for grid-like data using `DeepSaki.layers.ConvBlock2D()`.

    Architecture:
        ```mermaid
        flowchart LR
            i([input_tensor])-->up1 & up2
            subgraph ResBlockUp
                up1[UpSample2D]-->c1
                subgraph Block[2x]
                c1[ds.layers.Conv2DBlock]
                end
                up2[UpSample2D]-->c2[ds.layers.Conv2DBlock]
                c1 & c2-->add((+))
            end
            add-->o([output_tensor])
        ```

    """

    def __init__(
        self,
        activation: str = "leaky_relu",
        use_spec_norm: bool = False,
        use_bias: bool = True,
        padding: PaddingType = PaddingType.ZERO,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `ResBlockUp` layer.

        Args:
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults
                to False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(ResBlockUp, self).__init__()
        self.activation = activation
        self.use_spec_norm = use_spec_norm
        self.use_bias = use_bias
        self.padding = padding
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(ResBlockUp, self).build(input_shape)
        self.convRes = Conv2DBlock(
            filters=input_shape[-1],
            kernels=1,
            split_kernels=False,
            number_of_blocks=1,
            activation=self.activation,
            use_spec_norm=self.use_spec_norm,
            use_bias=self.use_bias,
            padding=self.padding,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )
        self.conv1 = Conv2DBlock(
            filters=input_shape[-1],
            kernels=3,
            split_kernels=False,
            number_of_blocks=1,
            activation=self.activation,
            use_spec_norm=self.use_spec_norm,
            use_bias=self.use_bias,
            padding=self.padding,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )
        self.conv2 = Conv2DBlock(
            filters=input_shape[-1],
            kernels=3,
            split_kernels=False,
            number_of_blocks=1,
            activation=self.activation,
            use_spec_norm=self.use_spec_norm,
            use_bias=self.use_bias,
            padding=self.padding,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `ResBlockUp` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (b,h,w,c)

        Returns:
            Tensor of shape (b,h*2,w*2,c)
        """
        path1 = inputs
        path2 = inputs

        path1 = tf.keras.layers.UpSampling2D()(path1)
        path1 = self.convRes(path1)

        path2 = tf.keras.layers.UpSampling2D()(path2)
        path2 = self.conv1(path2)
        path2 = self.conv2(path2)

        return tf.keras.layers.Add()([path1, path2])

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(ResBlockUp, self).get_config()
        config.update(
            {
                "activation": self.activation,
                "use_spec_norm": self.use_spec_norm,
                "use_bias": self.use_bias,
                "padding": self.padding,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


@tf.keras.utils.register_keras_serializable(package="Custom", name="scale")
class ScaleLayer(tf.keras.layers.Layer):
    """Trainable non-negative scalar that might be used as a trainable gate.

    It is a single learnable weight that is multiplied by all weights of the input tensor through broadcasting.
    """

    def __init__(self, initializer: Optional[tf.keras.initializers.Initializer] = None) -> None:
        """Initializes the `ScaleLayer` layer.

        Args:
            initializer (tf.keras.initializers.Initializer, optional): Initializer used to initialize the scalar weight.
                Defaults to None..
        """
        super(ScaleLayer, self).__init__()
        self.initializer = tf.keras.initializers.Ones() if initializer is None else initializer
        self.scale = self.add_weight(shape=[1], initializer=initializer, constraint=NonNegative(), trainable=True)
        self.input_spec = tf.keras.layers.InputSpec(min_ndim=1)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `ScaleLayer` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (batch, ...).

        Returns:
            Non-negative scaled version of the input. Tensor of shape (batch, ...).
        """
        return inputs * self.scale

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(ScaleLayer, self).get_config()
        config.update({"scale": self.scale, "initializer": self.initializer})
        return config


class ScalarGatedSelfAttention(tf.keras.layers.Layer):
    """Scaled dot-product self attention that is gated by a learnable scalar.

    Info:
        Implementation as used in the [VoloGAN paper](https://arxiv.org/abs/2207.09204).

    Architecture:
        ```mermaid
        flowchart LR
            i([input_tensor])-->f & g & h
            subgraph ScalarGatedSelfAttention
                f[ds.layers.DenseBlock] --> t[Transpose]
                g[ds.layers.DenseBlock] & t --> m1[Multiply] --> s[SoftMax]
                h[ds.layers.DenseBlock] & s --> m2[Multiply] --> v[DenseBlock] --> sc[ds.layers.ScaleLayer]
            end
            sc -->o([output_tensor])
        ```
    """

    def __init__(
        self,
        use_spec_norm: bool = False,
        intermediate_channel: Optional[bool] = None,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `ScalarGatedSelfAttention` layer.

        Args:
            use_spec_norm (bool, optional): If true, applies spectral normalization to convolutional and dense layers.
                Defaults to False.
            intermediate_channel (Optional[bool], optional): Intermediate channels for the self attention mechanism.
                Defaults to None.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the kernel weights.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(ScalarGatedSelfAttention, self).__init__()
        self.use_spec_norm = use_spec_norm
        self.intermediate_channel = intermediate_channel
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(ScalarGatedSelfAttention, self).build(input_shape)
        num_channel = input_shape[-1]
        if self.intermediate_channel is None:
            self.intermediate_channel = num_channel

        self.w_f = DenseBlock(
            units=self.intermediate_channel,
            use_spec_norm=self.use_spec_norm,
            number_of_blocks=1,
            activation=None,
            apply_final_normalization=False,
            use_bias=False,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )
        self.w_g = DenseBlock(
            units=self.intermediate_channel,
            use_spec_norm=self.use_spec_norm,
            number_of_blocks=1,
            activation=None,
            apply_final_normalization=False,
            use_bias=False,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )
        self.w_h = DenseBlock(
            units=self.intermediate_channel,
            use_spec_norm=self.use_spec_norm,
            number_of_blocks=1,
            activation=None,
            apply_final_normalization=False,
            use_bias=False,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )
        self.w_fgh = DenseBlock(
            units=num_channel,
            use_spec_norm=self.use_spec_norm,
            number_of_blocks=1,
            activation=None,
            apply_final_normalization=False,
            use_bias=False,
            kernel_initializer=self.kernel_initializer,
            gamma_initializer=self.gamma_initializer,
        )

        self.LN_f = tf.keras.layers.LayerNormalization(gamma_initializer=self.gamma_initializer)
        self.LN_g = tf.keras.layers.LayerNormalization(gamma_initializer=self.gamma_initializer)
        self.LN_h = tf.keras.layers.LayerNormalization(gamma_initializer=self.gamma_initializer)
        self.LN_fgh = tf.keras.layers.LayerNormalization(gamma_initializer=self.gamma_initializer)
        self.scale = ScaleLayer()

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `ScalarGatedSelfAttention` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (batch, height, width, channel)

        Returns:
            Tensor with same shape as input.
        """
        f = self.w_f(inputs)
        f = self.LN_f(f)
        f = tf.keras.layers.Permute(dims=(2, 1, 3))(f)

        g = self.w_g(inputs)
        g = self.LN_g(g)

        h = self.w_h(inputs)
        h = self.LN_h(h)

        f_g = tf.keras.layers.Multiply()([f, g])
        f_g = tf.keras.layers.Softmax(axis=1)(f_g)

        f_g_h = tf.keras.layers.Multiply()([f_g, h])
        f_g_h = self.w_fgh(f_g_h)
        f_g_h = self.LN_fgh(f_g_h)
        f_g_h = self.scale(f_g_h)

        return tf.keras.layers.Add()([f_g_h, inputs])

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(ScalarGatedSelfAttention, self).get_config()
        config.update(
            {
                "use_spec_norm": self.use_spec_norm,
                "intermediate_channel": self.intermediate_channel,
                "gamma_initializer": self.gamma_initializer,
                "kernel_initializer": self.kernel_initializer,
            }
        )
        return config



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
from contextlib import nullcontext as does_not_raise

import tensorflow as tf

from DeepSaki.layers.padding import ReflectionPadding2D
from tests.DeepSaki_test.layers_test.layers_test import CommonLayerChecks
from tests.DeepSaki_test.layers_test.layers_test import DeepSakiLayerChecks


class TestReflectionPadding2D(DeepSakiLayerChecks):
    @pytest.fixture()
    def reflection_padding_2d(self):
        return ReflectionPadding2D()

    @pytest.mark.parametrize(
        ("input", "padding", "expected_shape"),
        [
            (tf.ones(shape=(8, 64, 64, 3)), (0, 0), tf.TensorShape((8, 64, 64, 3))),
            (tf.ones(shape=(8, 64, 64, 3)), (1, 1), tf.TensorShape((8, 66, 66, 3))),
            (tf.ones(shape=(8, 64, 64, 3)), (2, 2), tf.TensorShape((8, 68, 68, 3))),
            (tf.ones(shape=(8, 64, 64, 3)), (2, 1), tf.TensorShape((8, 68, 66, 3))),
            (tf.ones(shape=(8, 64, 64, 3)), (0, 1), tf.TensorShape((8, 64, 66, 3))),
        ],
    )
    def test_padding_func_correct_shape(self, input, padding, expected_shape):
        layer = ReflectionPadding2D(padding=padding)
        output = layer._padding_func(input)
        assert output.shape == expected_shape

    @pytest.mark.parametrize(
        ("input", "padding", "expected_shape"),
        [
            (tf.TensorShape((8, 64, 64, 3)), (0, 0), tf.TensorShape((8, 64, 64, 3))),
            (tf.TensorShape((8, 64, 64, 3)), (1, 1), tf.TensorShape((8, 66, 66, 3))),
            (tf.TensorShape((8, 64, 64, 3)), (2, 2), tf.TensorShape((8, 68, 68, 3))),
            (tf.TensorShape((8, 64, 64, 3)), (2, 1), tf.TensorShape((8, 68, 66, 3))),
            (tf.TensorShape((8, 64, 64, 3)), (0, 1), tf.TensorShape((8, 64, 66, 3))),
        ],
    )
    def test_compute_outputshape_correct_shape(self, input, padding, expected_shape):
        layer = ReflectionPadding2D(padding=padding)
        output = layer.compute_output_shape(input)
        assert output == expected_shape

    @pytest.mark.parametrize(
        ("input_shape", "padding", "expected_shape"),
        [
            (tf.TensorShape((8, 64, 64, 3)), (0, 0), tf.TensorShape((8, 64, 64, 3))),
            (tf.TensorShape((8, 64, 64, 3)), (1, 1), tf.TensorShape((8, 66, 66, 3))),
            (tf.TensorShape((8, 64, 64, 3)), (2, 2), tf.TensorShape((8, 68, 68, 3))),
            (tf.TensorShape((8, 64, 64, 3)), (2, 1), tf.TensorShape((8, 68, 66, 3))),
            (tf.TensorShape((8, 64, 64, 3)), (0, 1), tf.TensorShape((8, 64, 66, 3))),
        ],
    )
    def test_call_correct_output_shape(self, input_shape, padding, expected_shape):
        layer_instance = ReflectionPadding2D(padding=padding)
        CommonLayerChecks.has_call_correct_output_shape(layer_instance, input_shape, expected_shape)

    @pytest.mark.parametrize(
        ("input_shape", "expected_context"),
        [
            (tf.TensorShape((8, 64, 64, 4)), does_not_raise()),
            (tf.TensorShape((1, 32, 32, 3)), does_not_raise()),
            (tf.TensorShape((8, 64, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64)), pytest.raises(ValueError)),
            (tf.TensorShape((8)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5)), pytest.raises(ValueError)),
            (tf.TensorShape((8, 64, 64, 4, 5, 6)), pytest.raises(ValueError)),
        ],
    )
    def test_call_raises_error_wrong_input_spec(self, reflection_padding_2d, input_shape, expected_context):
        CommonLayerChecks.does_call_raises_error_wrong_input_spec(reflection_padding_2d, input_shape, expected_context)



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import tensorflow as tf

from DeepSaki.layers.layer_composites import Conv2DBlock
from DeepSaki.layers.layer_composites import DownSampleBlock
from DeepSaki.layers.layer_composites import ResBlockDown
from DeepSaki.layers.layer_composites import ResBlockUp
from DeepSaki.layers.layer_composites import ResidualBlock
from DeepSaki.layers.layer_composites import ScalarGatedSelfAttention
from DeepSaki.layers.layer_composites import UpSampleBlock
from DeepSaki.layers.layer_helper import dropout_func
from DeepSaki.types.layers_enums import DownSampleType
from DeepSaki.types.layers_enums import PaddingType
from DeepSaki.types.layers_enums import UpSampleType

class Encoder(tf.keras.layers.Layer):
    """Combines conv blocks with down sample blocks.

    The spatial width is halved with every level while the channel depth is doubled.
    Can be combined with `ds.layers.Decoder` and `ds.layers.Bottleneck` to form an auto encoder model.

    Tipp:
        Checkout the ds.models api to find models using this layer.
    """

    def __init__(
        self,
        number_of_levels: int = 3,
        filters: int = 64,
        limit_filters: int = 1024,
        downsampling: DownSampleType = DownSampleType.CONV_STRIDE_2,
        kernels: int = 3,
        split_kernels: bool = False,
        number_of_blocks: int = 2,
        activation: str = "leaky_relu",
        first_kernel: Optional[int] = None,
        use_ResidualBlock: bool = False,
        residual_cardinality: int = 1,
        channel_list: Optional[List[int]] = None,
        use_spec_norm: bool = False,
        use_bias: bool = True,
        dropout_rate: float = 0.0,
        use_self_attention: bool = False,
        omit_skips: int = 0,
        padding: PaddingType = PaddingType.ZERO,
        output_skips: bool = False,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `Encoder` layer.

        Args:
            number_of_levels (int, optional): Number of downsampling levels of the model. Defaults to 3.
            filters (int, optional): Number of filters for the initial encoder block. Defaults to 64.
            limit_filters (int, optional): Limits the number of filters, which is doubled with every downsampling block.
                Defaults to 1024.
            downsampling (DownSampleType, optional): Describes the downsampling method. Defaults to DownSampleType.CONV_STRIDE_2.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)` and
                `(1,kernel)` respectivly. Defaults to False.
            number_of_blocks (int, optional): Number of consecutive convolutional building blocks, i.e. `Conv2DBlock`.
                Defaults to 2.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            first_kernel (Optional[int], optional): The first convolution can have a different kernel size, to e.g.
                increase the perceptive field, while the channel depth is still low. Defaults to None.
            use_ResidualBlock (bool, optional): Whether or not to use the `ResidualBlock` instead of the
                `Conv2DBlock`. Defaults to False.
            residual_cardinality (int, optional): Cardinality for the `ResidualBlock`. Defaults to 1.
            channel_list (Optional[List[int]], optional): alternativly to number_of_layers and filters, a list with the
                disired filters for each level can be provided. e.g. channel_list = [64, 128, 256] results in a 3-level
                Encoder with 64, 128, 256 filters for level 1, 2 and 3 respectivly. Defaults to None.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers.
                Defaults to False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.0.
            use_self_attention (bool, optional): Determines whether to apply self-attention in the encoder. Defaults to False.
            omit_skips (int, optional): Defines how many layers should not output a skip connection output. Requires
                `output_skips` to be True. E.g. if `omit_skips = 2`, the first two levels do not output a skip connection,
                it starts at level 3. Defaults to 0.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            output_skips (bool, optional): If true, ski connections are output. Could be used to attach an encoder to a
                decoder model. Defaults to False.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(Encoder, self).__init__()
        self.number_of_levels = number_of_levels
        self.filters = filters
        self.limit_filters = limit_filters
        self.downsampling = downsampling
        self.kernels = kernels
        self.split_kernels = split_kernels
        self.number_of_blocks = number_of_blocks
        self.activation = activation
        self.first_kernel = first_kernel
        self.use_ResidualBlock = use_ResidualBlock
        self.residual_cardinality = residual_cardinality
        self.channel_list = channel_list
        self.use_spec_norm = use_spec_norm
        self.dropout_rate = dropout_rate
        self.use_self_attention = use_self_attention
        self.omit_skips = omit_skips
        self.padding = padding
        self.output_skips = output_skips
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer

        self.input_spec = tf.keras.layers.InputSpec(ndim=4, dtype=tf.float32)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(Encoder, self).build(input_shape)

        if self.channel_list is None:
            self.channel_list = [min(self.filters * 2**i, self.limit_filters) for i in range(self.number_of_levels)]
        else:
            self.number_of_levels = len(self.channel_list)

        self.encoderBlocks = []
        self.downSampleBlocks = []

        if self.use_self_attention:
            self.SA = ScalarGatedSelfAttention(
                use_spec_norm=self.use_spec_norm,
                intermediate_channel=None,
                kernel_initializer=self.kernel_initializer,
                gamma_initializer=self.gamma_initializer,
            )
        else:
            self.SA = None

        for i, ch in enumerate(self.channel_list):
            encoder_kernels = self.first_kernel if i == 0 and self.first_kernel else self.kernels

            if self.use_ResidualBlock:
                self.encoderBlocks.append(
                    ResidualBlock(
                        filters=ch,
                        activation=self.activation,
                        kernels=encoder_kernels,
                        number_of_blocks=self.number_of_blocks,
                        use_spec_norm=self.use_spec_norm,
                        dropout_rate=self.dropout_rate,
                        use_bias=self.use_bias,
                        residual_cardinality=self.residual_cardinality,
                        padding=self.padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )
                self.downSampleBlocks.append(
                    ResBlockDown(
                        activation=self.activation,
                        use_spec_norm=self.use_spec_norm,
                        use_bias=self.use_bias,
                        padding=self.padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )
            else:
                self.encoderBlocks.append(
                    Conv2DBlock(
                        filters=ch,
                        kernels=encoder_kernels,
                        split_kernels=self.split_kernels,
                        activation=self.activation,
                        number_of_blocks=self.number_of_blocks,
                        use_spec_norm=self.use_spec_norm,
                        dropout_rate=self.dropout_rate,
                        padding=self.padding,
                        use_bias=self.use_bias,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )
                self.downSampleBlocks.append(
                    DownSampleBlock(
                        downsampling=self.downsampling,
                        activation=self.activation,
                        kernels=encoder_kernels,
                        use_spec_norm=self.use_spec_norm,
                        padding=self.padding,
                        use_bias=self.use_bias,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )

    def call(self, inputs: tf.Tensor) -> Union[tf.Tensor, Tuple[tf.Tensor, tf.Tensor]]:
        """Calls the `Encoder` layer.

        Args:
            inputs (tf.Tensor): Input tensor of shape (batch, height, width, channel)

        Returns:
            If `output_skips=False` only the final output of the Encoder is returned as a tensor of shape
                (`batch`, `height/2**number_of_levels`, `width/2**number_of_levels`,
                `min(filters * 2 ** (number_of_levels - 1), limit_filters)`. If `output_skips=True` additionally returns
                a tensor of tensor (one for each level of the encoder) of shape (`batch`, `height/2**level`,
                `width/2**level`, `min(filters * 2 ** (number_of_levels - 1), limit_filters)`.
        """
        x = inputs
        skips = []

        for level in range(self.number_of_levels):
            if level == 3 and self.SA is not None:
                x = self.SA(x)
            skip = self.encoderBlocks[level](x)
            x = self.downSampleBlocks[level](skip)
            if self.output_skips:
                if level >= self.omit_skips:  # omit the first skip connection
                    skips.append(skip)
                else:
                    skips.append(None)

        if self.output_skips:
            return x, skips
        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(Encoder, self).get_config()
        config.update(
            {
                "number_of_levels": self.number_of_levels,
                "filters": self.filters,
                "limit_filters": self.limit_filters,
                "downsampling": self.downsampling,
                "kernels": self.kernels,
                "split_kernels": self.split_kernels,
                "number_of_blocks": self.number_of_blocks,
                "activation": self.activation,
                "first_kernel": self.first_kernel,
                "use_ResidualBlock": self.use_ResidualBlock,
                "residual_cardinality": self.residual_cardinality,
                "channel_list": self.channel_list,
                "use_spec_norm": self.use_spec_norm,
                "use_bias": self.use_bias,
                "dropout_rate": self.dropout_rate,
                "use_self_attention": self.use_self_attention,
                "omit_skips": self.omit_skips,
                "padding": self.padding,
                "output_skips": self.output_skips,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class Bottleneck(tf.keras.layers.Layer):
    """Bottlenecks are sub-model blocks in auto-encoder-like models such as UNet or ResNet.

    It is composed of multiple convolution blocks which might have residuals.

    Can be combined with `ds.layers.Encoder` and `ds.layers.Decoder` to form an auto encoder model.

    Tipp:
        Checkout the ds.models api to find models using this layer.

    """

    def __init__(
        self,
        n_bottleneck_blocks: int = 3,
        kernels: int = 3,
        split_kernels: bool = False,
        number_of_blocks: int = 2,
        use_ResidualBlock: bool = False,
        activation: str = "leaky_relu",
        dropout_rate: float = 0.2,
        channel_list: Optional[List[int]] = None,
        use_spec_norm: bool = False,
        use_bias: bool = True,
        residual_cardinality: int = 1,
        padding: PaddingType = PaddingType.ZERO,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `Bottleneck` layer.

        Args:
            n_bottleneck_blocks (int, optional): Number of consecutive blocks. Defaults to 3.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)`
                and `(1,kernel)` respectivly. Defaults to False.
            number_of_blocks (int, optional): : Number of consecutive conv layers within a basic building block.
                Defaults to 2.
            use_ResidualBlock (bool, optional): Whether or not to use the `ResidualBlock` instead of the
                `Conv2DBlock`. Defaults to False.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.2.
            channel_list (Optional[List[int]], optional): Alternativly to number_of_layers and filters, a list with the
                disired filters for each level can be provided. e.g. channel_list = [64, 128, 256] results in a 3-level
                Decoder with 64, 128, 256 filters for level 1, 2 and 3 respectivly. Defaults to None.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults
                to False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            residual_cardinality (int, optional): Cardinality for the `ResidualBlock`. Defaults to 1.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(Bottleneck, self).__init__()
        self.use_ResidualBlock = use_ResidualBlock
        self.n_bottleneck_blocks = n_bottleneck_blocks
        self.kernels = kernels
        self.split_kernels = split_kernels
        self.number_of_blocks = number_of_blocks
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.channel_list = channel_list
        self.use_spec_norm = use_spec_norm
        self.use_bias = use_bias
        self.residual_cardinality = residual_cardinality
        self.padding = padding
        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer
        self.input_spec = tf.keras.layers.InputSpec(ndim=4, dtype=tf.float32)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(Bottleneck, self).build(input_shape)

        if self.channel_list is None:
            ch = input_shape[-1]
            self.channel_list = [ch for _ in range(self.n_bottleneck_blocks)]

        self.layers = []
        for ch in self.channel_list:
            if self.use_ResidualBlock:
                self.layers.append(
                    ResidualBlock(
                        activation=self.activation,
                        filters=ch,
                        kernels=self.kernels,
                        number_of_blocks=self.number_of_blocks,
                        use_spec_norm=self.use_spec_norm,
                        use_bias=self.use_bias,
                        residual_cardinality=self.residual_cardinality,
                        padding=self.padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )
            else:
                self.layers.append(
                    Conv2DBlock(
                        filters=ch,
                        kernels=self.kernels,
                        split_kernels=self.split_kernels,
                        number_of_blocks=self.number_of_blocks,
                        activation=self.activation,
                        use_spec_norm=self.use_spec_norm,
                        use_bias=self.use_bias,
                        padding=self.padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )

        self.dropout = dropout_func(self.channel_list[-1], self.dropout_rate)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `Bottleneck` layer.

        Args:
            inputs (tf.Tensor): Input tensor of shape (batch, height, width, channel)

        Returns:
            Tensor of shape (batch, height, width, channel)
        """
        x = inputs

        for layer in self.layers:
            x = layer(x)

        if self.dropout_rate > 0:
            x = self.dropout(x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(Bottleneck, self).get_config()
        config.update(
            {
                "use_ResidualBlock": self.use_ResidualBlock,
                "n_bottleneck_blocks": self.n_bottleneck_blocks,
                "kernels": self.kernels,
                "split_kernels": self.split_kernels,
                "number_of_blocks": self.number_of_blocks,
                "activation": self.activation,
                "dropout_rate": self.dropout_rate,
                "use_spec_norm": self.use_spec_norm,
                "use_bias": self.use_bias,
                "channel_list": self.channel_list,
                "residual_cardinality": self.residual_cardinality,
                "padding": self.padding,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config


class Decoder(tf.keras.layers.Layer):
    """Combines conv blocks with up sample blocks.

    The spatial width is doubled with every level while the channel depth is halfed.
    Can be combined with `ds.layers.Encoder` and `ds.layers.Bottleneck` to form an auto encoder model.

    Tipp:
        Checkout the ds.models api to find models using this layer.
    """

    def __init__(
        self,
        number_of_levels: int = 3,
        upsampling: UpSampleType = UpSampleType.RESAMPLE_AND_CONV,
        filters: int = 64,
        limit_filters: int = 1024,
        kernels: int = 3,
        split_kernels: bool = False,
        number_of_blocks: int = 2,
        activation: str = "leaky_relu",
        dropout_rate: float = 0.2,
        use_ResidualBlock: bool = False,
        residual_cardinality: int = 1,
        channel_list: Optional[List[int]] = None,
        use_spec_norm: bool = False,
        use_bias: bool = True,
        use_self_attention: bool = False,
        enable_skip_connections_input: bool = False,
        padding: PaddingType = PaddingType.ZERO,
        kernel_initializer: Optional[tf.keras.initializers.Initializer] = None,
        gamma_initializer: Optional[tf.keras.initializers.Initializer] = None,
    ) -> None:
        """Initializes the `Decoder` layer.

        Args:
            number_of_levels (int, optional): Number levels in the decoder. Effectivly the number of convolution and
                upsample pairs. Defaults to 3.
            upsampling (UpSampleType, optional): Describes the upsampling method used. Defaults to
                UpSampleType.RESAMPLE_AND_CONV.
            filters (int, optional): Base size of filters the is doubled with every level of the decoder.
                Defaults to 64.
            limit_filters (int, optional): Limits the number of filters, which is doubled with every downsampling block.
                Defaults to 512.
            kernels (int, optional): Size of the convolutions kernels. Defaults to 3.
            split_kernels (bool, optional): To decrease the number of parameters, a convolution with the kernel_size
                `(kernel,kernel)` can be splitted into two consecutive convolutions with the kernel_size `(kernel,1)` and
                `(1,kernel)` respectivly. Defaults to False.
            number_of_blocks (int, optional): Number of consecutive convolutional building blocks, i.e. `Conv2DBlock`.
                Defaults to 2.
            activation (str, optional): String literal or tensorflow activation function object to obtain activation
                function. Defaults to "leaky_relu".
            dropout_rate (float, optional): Probability of the dropout layer. If the preceeding layer has more than one
                channel, spatial dropout is applied, otherwise standard dropout. Defaults to 0.2.
            use_ResidualBlock (bool, optional): Whether or not to use the `ResidualBlock` instead of the
                `Conv2DBlock`. Defaults to False.
            residual_cardinality (int, optional): Cardinality for the `ResidualBlock`. Defaults to 1.
            channel_list (Optional[List[int]], optional): Alternativly to number_of_layers and filters, a list with the
                disired filters for each level can be provided. e.g. channel_list = [64, 128, 256] results in a 3-level
                Decoder with 64, 128, 256 filters for level 1, 2 and 3 respectivly. Defaults to None.
            use_spec_norm (bool, optional): Applies spectral normalization to convolutional and dense layers. Defaults to False.
            use_bias (bool, optional): Whether convolutions and dense layers include a bias or not. Defaults to True.
            use_self_attention (bool, optional): Determines whether to apply self-attention in the decoder. Defaults to False.
            enable_skip_connections_input (bool, optional): Whether or not to input skip connections at each level. Defaults to False.
            padding (PaddingType, optional): Padding type. Defaults to PaddingType.ZERO.
            kernel_initializer (tf.keras.initializers.Initializer, optional): Initialization of the convolutions kernels.
                Defaults to None.
            gamma_initializer (tf.keras.initializers.Initializer, optional): Initialization of the normalization layers.
                Defaults to None.
        """
        super(Decoder, self).__init__()
        self.number_of_levels = number_of_levels
        self.filters = filters
        self.upsampling = upsampling
        self.limit_filters = limit_filters
        self.kernels = kernels
        self.split_kernels = split_kernels
        self.number_of_blocks = number_of_blocks
        self.activation = activation
        self.use_ResidualBlock = use_ResidualBlock
        self.channel_list = channel_list
        self.use_spec_norm = use_spec_norm
        self.use_bias = use_bias
        self.dropout_rate = dropout_rate
        self.use_self_attention = use_self_attention
        self.enable_skip_connections_input = enable_skip_connections_input
        self.residual_cardinality = residual_cardinality
        self.padding = padding

        self.kernel_initializer = kernel_initializer
        self.gamma_initializer = gamma_initializer

        self.input_spec = tf.keras.layers.InputSpec(ndim=4, dtype=tf.float32)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        super(Decoder, self).build(input_shape)

        if self.channel_list is None:
            self.channel_list = [
                min(self.filters * 2**i, self.limit_filters) for i in reversed(range(self.number_of_levels))
            ]
        else:
            self.number_of_levels = len(self.channel_list)

        self.decoderBlocks = []
        self.upSampleBlocks = []

        if self.use_self_attention:
            self.SA = ScalarGatedSelfAttention(
                use_spec_norm=self.use_spec_norm,
                intermediate_channel=None,
                kernel_initializer=self.kernel_initializer,
                gamma_initializer=self.gamma_initializer,
            )
        else:
            self.SA = None

        for i, ch in enumerate(self.channel_list):
            dropout_rate = self.dropout_rate if i < int(self.number_of_levels / 2) else 0

            if self.use_ResidualBlock:
                self.decoderBlocks.append(
                    ResidualBlock(
                        filters=ch,
                        activation=self.activation,
                        kernels=self.kernels,
                        number_of_blocks=self.number_of_blocks,
                        use_spec_norm=self.use_spec_norm,
                        dropout_rate=dropout_rate,
                        use_bias=self.use_bias,
                        residual_cardinality=self.residual_cardinality,
                        padding=self.padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )
                self.upSampleBlocks.append(
                    ResBlockUp(
                        activation=self.activation,
                        use_spec_norm=self.use_spec_norm,
                        use_bias=self.use_bias,
                        padding=self.padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )
            else:
                self.decoderBlocks.append(
                    Conv2DBlock(
                        filters=ch,
                        kernels=self.kernels,
                        split_kernels=self.split_kernels,
                        activation=self.activation,
                        number_of_blocks=self.number_of_blocks,
                        dropout_rate=dropout_rate,
                        use_spec_norm=self.use_spec_norm,
                        use_bias=self.use_bias,
                        padding=self.padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )
                self.upSampleBlocks.append(
                    UpSampleBlock(
                        kernels=self.kernels,
                        upsampling=self.upsampling,
                        activation=self.activation,
                        use_spec_norm=self.use_spec_norm,
                        use_bias=self.use_bias,
                        padding=self.padding,
                        kernel_initializer=self.kernel_initializer,
                        gamma_initializer=self.gamma_initializer,
                    )
                )

    def call(self, inputs: Union[tf.Tensor, Tuple[tf.Tensor, tf.Tensor]]) -> tf.Tensor:
        """Calls the `Decoder` layer.

        Args:
            inputs (Union[tf.Tensor , Tuple[tf.Tensor, tf.Tensor]]): If `enable_skip_connections_input=False` only
                inputs a tensor of shape (batch, height, width, min(channel*2**level, limit_filters)). If
                `enable_skip_connections_input=True`, additonally at every level of the decoder, skip connections from
                an encoder can be inserted.

        Returns:
            tf.Tensor: Tensor of shape (`batch`, `height*2**number_of_levels`, `width*2**number_of_levels`,`filters`).
        """
        skip_connections = None
        if self.enable_skip_connections_input:
            x, skip_connections = inputs
        else:
            x = inputs

        for level in range(self.number_of_levels):
            if level == 3 and self.SA is not None:
                x = self.SA(x)
            x = self.upSampleBlocks[level](x)
            if skip_connections is not None:
                x = tf.keras.layers.concatenate([x, skip_connections[self.number_of_levels - (level + 1)]])
            x = self.decoderBlocks[level](x)

        return x

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(Decoder, self).get_config()
        config.update(
            {
                "number_of_levels": self.number_of_levels,
                "filters": self.filters,
                "limit_filters": self.limit_filters,
                "upsampling": self.upsampling,
                "kernels": self.kernels,
                "split_kernels": self.split_kernels,
                "number_of_blocks": self.number_of_blocks,
                "activation": self.activation,
                "use_ResidualBlock": self.use_ResidualBlock,
                "residual_cardinality": self.residual_cardinality,
                "channel_list": self.channel_list,
                "use_spec_norm": self.use_spec_norm,
                "dropout_rate": self.dropout_rate,
                "use_bias": self.use_bias,
                "use_self_attention": self.use_self_attention,
                "enable_skip_connections_input": self.enable_skip_connections_input,
                "padding": self.padding,
                "kernel_initializer": self.kernel_initializer,
                "gamma_initializer": self.gamma_initializer,
            }
        )
        return config



from typing import Any
from typing import Dict

import tensorflow as tf

class GlobalSumPooling2D(tf.keras.layers.Layer):
    """Global sum pooling operation for spatial data.

    Tips:
        Similar to tensorflow's [GlobalMaxPooling2D](https://www.tensorflow.org/api_docs/python/tf/keras/layers/GlobalMaxPooling2D)
        and [GlobalAveragePooling2D](https://www.tensorflow.org/api_docs/python/tf/keras/layers/GlobalAveragePooling2D)
    """

    def __init__(self, data_format: str = "channels_last", **kwargs: Any) -> None:
        """Initialize the `GlobalSumPooling2D` object.

        Raises:
            ValueError: If `data_format` is not supported.

        Args:
            data_format (str, optional): ["channels_last"|"data_format"]
            kwargs (Any): Additional key word arguments passed to the base class.
        """
        super(GlobalSumPooling2D, self).__init__(**kwargs)
        match data_format:
            case "channels_last" | "channels_first":
                self.data_format = data_format
            case _:
                raise ValueError("Unsupported channel configuration provided")
        self.channel_axis = 3 if self.data_format == "channels_last" else 1
        self.axis_to_sum = (1, 2) if self.data_format == "channels_last" else (2, 3)
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

    def compute_output_shape(self, input_shape: tf.TensorShape) -> tf.TensorShape:
        """Computes the output shape of the layer.

        This method will cause the layer's state to be built, if that has not happened before. This requires that the
        layer will later be used with inputs that match the input shape provided here.

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.

        Returns:
            A TensorShape instance representing the shape of the layer's output Tensor.
        """
        input_shape = tf.TensorShape(input_shape).as_list()
        return tf.TensorShape([input_shape[0], input_shape[self.channel_axis]])

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        """Calls the `GlobalSumPooling2D` layer.

        Args:
            inputs (tf.Tensor): Tensor of shape (`batch`,`height`,`width`,`channel`) or
                (`batch`,`channel`,`height`,`width`).

        Returns:
            Tensor of shape (`batch`,`channel`) where the elements are summed to reduce the axis.
        """
        return tf.reduce_sum(input_tensor=inputs, axis=self.axis_to_sum, keepdims=False)

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(GlobalSumPooling2D, self).get_config()
        config.update({"data_format": self.data_format})
        return config


class LearnedPooling(tf.keras.layers.Layer):
    """Layer that learns a pooling operation.

    Instead of using [MaxPooling2D](https://www.tensorflow.org/api_docs/python/tf/keras/layers/MaxPooling2D) or
    [AveragePooling2D](https://www.tensorflow.org/api_docs/python/tf/keras/layers/AveragePooling2D) the layer learns a
    pooling operation.

    Info:
        Under the hood this layer simply performs a non-overlapping convolution operation.
    """

    def __init__(self, pool_size: int = 2, **kwargs: Any) -> None:
        """Initializes an instance of `LearnedPooling`.

        Args:
            pool_size (int, optional): Size of the pooling window and the stride of the convolution operation. If the
                input is of shape (b, h, w, c), the output is (b,h/pool_size,w/pool_size,c). Defaults to 2.
            kwargs (Any): Additional key word arguments passed to the base class.
        """
        super(LearnedPooling, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.input_spec = tf.keras.layers.InputSpec(ndim=4)

    def build(self, input_shape: tf.TensorShape) -> None:
        """Build layer depending on the `input_shape` (output shape of the previous layer).

        Args:
            input_shape (tf.TensorShape): Shape of the input tensor to this layer.
        """
        self.pooling = tf.keras.layers.Conv2D(
            kernel_size=self.pool_size,
            strides=self.pool_size,
            filters=input_shape[-1],
            use_bias=False,
            padding="same",
        )

    def call(self, x: tf.Tensor) -> tf.Tensor:
        """Calls the `LearnedPooling` layer.

        Args:
            x (tf.Tensor): Tensor of shape (`batch`,`height`,`width`,`channel`).

        Returns:
            Pooled tensor of shape (`batch`,`height/pool_size`,`width/pool_size`,`channel`)
        """
        return self.pooling(x)

    def get_config(self) -> Dict[str, Any]:
        """Serialization of the object.

        Returns:
            Dictionary with the class' variable names as keys.
        """
        config = super(LearnedPooling, self).get_config()
        config.update({"pool_size": self.pool_size})
        return config



# pooling.py
from DeepSaki.layers.pooling import GlobalSumPooling2D
from DeepSaki.layers.pooling import LearnedPooling

# fourier_layer.py
from DeepSaki.layers.fourier_layer import FourierConvolution2D
from DeepSaki.layers.fourier_layer import FourierFilter2D
from DeepSaki.layers.fourier_layer import FFT2D
from DeepSaki.layers.fourier_layer import iFFT2D
from DeepSaki.layers.fourier_layer import FFT3D
from DeepSaki.layers.fourier_layer import iFFT3D
from DeepSaki.layers.fourier_layer import FourierPooling2D
from DeepSaki.layers.fourier_layer import rFFT2DFilter

# padding.py
from DeepSaki.layers.padding import ReflectionPadding2D

# layer_composites.py
from DeepSaki.layers.layer_composites import Conv2DSplitted
from DeepSaki.layers.layer_composites import Conv2DBlock
from DeepSaki.layers.layer_composites import DenseBlock
from DeepSaki.layers.layer_composites import DownSampleBlock
from DeepSaki.layers.layer_composites import UpSampleBlock
from DeepSaki.layers.layer_composites import ResidualBlock
from DeepSaki.layers.layer_composites import ResBlockDown
from DeepSaki.layers.layer_composites import ResBlockUp
from DeepSaki.layers.layer_composites import ScaleLayer
from DeepSaki.layers.layer_composites import ScalarGatedSelfAttention

# layer_composites.py
from DeepSaki.layers.sub_model_composites import Encoder
from DeepSaki.layers.sub_model_composites import Bottleneck
from DeepSaki.layers.sub_model_composites import Decoder

# layer_helper.py
# Fuctions
from DeepSaki.layers.layer_helper import plot_layer
from DeepSaki.layers.layer_helper import get_initializer
from DeepSaki.layers.layer_helper import pad_func
from DeepSaki.layers.layer_helper import dropout_func

__all__ = [
    "GlobalSumPooling2D",
    "LearnedPooling",
    "FourierFilter2D",
    "FFT2D",
    "iFFT2D",
    "FFT3D",
    "iFFT3D",
    "FourierConvolution2D",
    "FourierPooling2D",
    "rFFT2DFilter",
    "ReflectionPadding2D",
    "Conv2DSplitted",
    "Conv2DBlock",
    "DenseBlock",
    "DownSampleBlock",
    "ResBlockDown",
    "ResBlockUp",
    "UpSampleBlock",
    "ScaleLayer",
    "ScalarGatedSelfAttention",
    "Encoder",
    "ResidualBlock",
    "Bottleneck",
    "Decoder",
    "get_initializer",
    "plot_layer",
    "pad_func",
    "dropout_func",
]



from enum import Enum
from enum import auto

class LossCalcType(Enum):
    """`Enum` to define the loss calculation type used for image based losses derived from `ImageBasedLoss`.

    Attributes:
        PER_IMAGE: Loss is calculated over the entire image.
        PER_CHANNEL: Loss is calculated per channel.

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.losses.StructuralSimilarityLoss(
                calculation_type = ds.types.LossCalcType.PER_IMAGE
            )
    ```
    """

    PER_IMAGE = auto()
    PER_CHANNEL = auto()


class LossType(Enum):
    """`Enum` to define the loss type used for image based losses derived from `ImageBasedLoss`.

    Attributes:
        MAE: Loss is a form of mean absolute error.
        MSE: Loss is a form of mean squared error.

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.losses.PixelDistanceLoss(
                loss_type = ds.types.LossCalcType.MSE
            )
    ```
    """

    MAE = auto()
    MSE = auto()



from enum import Enum
from enum import auto

class MultiplicationType(Enum):
    """`Enum` used to define how two matrices shall be multiplied.

    Attributes:
        ELEMENT_WISE: Indicates to apply an element-wise multiplication of 2 tensors.
        MATRIX_PRODUCT: Indicates to apply a matrix-product between 2 tensors.

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.layers.FourierConvolution2D(
                multiplication_type = ds.types.MultiplicationType.MATRIX_PRODUCT
            )
    ```
    """

    ELEMENT_WISE = auto()
    MATRIX_PRODUCT = auto()


class FrequencyFilter(Enum):
    """`Enum` used to define valid filters for `rFFT2DFilter`.

    Attributes:
        LOW_PASS: Indicates that low frequency components shall be kept and high frequency components shall be
            filtered.
        HIGH_PASS: Indicates that high frequency components shall be kept and low frequency components shall be
            filtered.

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.layers.rFFT2DFilter(
                filter_type = ds.types.FrequencyFilter.LOW_PASS
            )
    ```
    """

    LOW_PASS = auto()
    HIGH_PASS = auto()


class PaddingType(Enum):
    """`Enum` used to define different types of padding opperations.

    Attributes:
        ZERO: Indicates to apply a zero padding operations.
        REFLECTION: Indicates to apply a reflection padding operation.

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.models.LayoutContentDiscriminator(
                padding = ds.types.PaddingType.ZERO
            )
    ```
    """

    NONE = auto()
    ZERO = auto()
    REFLECTION = auto()


class InitializerFunc(Enum):
    """`Enum` used to define different types of initializer functions.

    Attributes:
        RANDOM_NORMAL: Corresponds to a random normal initializer function.
        RANDOM_UNIFORM: Corresponds to a random uniform initializer function.
        GLOROT_NORMAL: Corresponds to a Glorot normal initializer function.
        GLOROT_UNIFORM: Corresponds to a Glorot uniform initializer function.
        HE_NORMAL: Corresponds to a He normal initializer function.
        HE_UNIFORM: Corresponds to a He uniform initializer function.
        HE_ALPHA_NORMAL: Corresponds to a He Alpha normal initializer function.
        HE_ALPHA_UNIFORM: Corresponds to a He Alpha Uniform initializer function.

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.layers.get_initializer(
                initializer = ds.types.InitializerFunc.HE_ALPHA_NORMAL
            )
    ```
    """

    NONE = auto()
    RANDOM_NORMAL = auto()
    RANDOM_UNIFORM = auto()
    GLOROT_NORMAL = auto()
    GLOROT_UNIFORM = auto()
    HE_NORMAL = auto()
    HE_UNIFORM = auto()
    HE_ALPHA_NORMAL = auto()
    HE_ALPHA_UNIFORM = auto()


class UpSampleType(Enum):
    """`Enum` used to define different types of up sampling functions.

    Attributes:
        RESAMPLE_AND_CONV: Use a resampling technique like interpolation followed by a convolution.
        TRANSPOSE_CONV: Use a transpose convolution.
        DEPTH_TO_SPACE: Use a depth to space operation after a convolution the adjust the filters.

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.layers.Decoder(
                upsampling = ds.types.UpSampleType.TRANSPOSE_CONV
            )
    ```
    """

    RESAMPLE_AND_CONV = auto()
    TRANSPOSE_CONV = auto()
    DEPTH_TO_SPACE = auto()


class DownSampleType(Enum):
    """`Enum` used to define different types of down sampling functions.

    Attributes:
        AVG_POOLING: Use average pooling.
        CONV_STRIDE_2: Use a convolution with a stride of (2,2).
        MAX_POOLING: Use max pooling.
        SPACE_TO_DEPTH: Use a space to depth operation followed by a convolution the adjust the filters.

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.layers.Encoder(
                downsampling = ds.types.DownSampleType.SPACE_TO_DEPTH
            )
    ```
    """

    AVG_POOLING = auto()
    CONV_STRIDE_2 = auto()
    MAX_POOLING = auto()
    SPACE_TO_DEPTH = auto()


class LinearLayerType(Enum):
    """`Enum` used to define different types of linear functions.

    Attributes:
        MLP: Uses an MLP.
        CONV_1x1: Use a convolution with a kernel of (1,1).

    **Example:**
    ```python hl_lines="3"
    import DeepSaki as ds
    layer = ds.models.UNet(
                linear_layer_type = ds.types.LinearLayerType.MLP
            )
    ```
    """

    MLP = auto()
    CONV_1x1 = auto()



import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # deactivate tensorflow warnings and infos. Keep Errors
import tensorflow as tf

from DeepSaki.constraints.constraints import NonNegative


class TestNonNegativeConstraint:
    @pytest.fixture(scope="class")
    def non_negative_constraint(self):
        return NonNegative()

    @pytest.mark.parametrize(
        "input_dtype",
        [
            tf.float16,
            tf.bfloat16,
            tf.float32,
            tf.float64,
        ],
    )
    def test_dtype_output_same_as_input(self, non_negative_constraint, input_dtype):
        output = non_negative_constraint(tf.constant([1, 2, 3], dtype=input_dtype))
        assert output.dtype == input_dtype

    @pytest.mark.parametrize(
        "input_tensor",
        [
            tf.constant([1.0]),
            tf.constant([1.0, 1.0]),
            tf.constant([[1.0, 1.0], [1.0, 1.0]]),
            tf.constant([[[1.0, 1.0], [1.0, 1.0]], [[1.0, 1.0], [1.0, 1.0]]]),
        ],
    )
    def test_shape_output_same_as_input(self, non_negative_constraint, input_tensor):
        output = non_negative_constraint(input_tensor)
        assert output.shape == input_tensor.shape

    @pytest.mark.parametrize(
        ("input", "expected"),
        [
            (tf.constant([1.0]), tf.constant([1.0])),
            (tf.constant([-1.0]), tf.constant([0.0])),
            (tf.constant([1.0, 1.0]), tf.constant([1.0, 1.0])),
            (tf.constant([1.0, -1.0]), tf.constant([1.0, 0.0])),
            (tf.constant([-1.0, -1.0]), tf.constant([0.0, 0.0])),
            (tf.constant([[1.0, 1.0], [1.0, 1.0]]), tf.constant([[1.0, 1.0], [1.0, 1.0]])),
            (tf.constant([[-1.0, -1.0], [1.0, 1.0]]), tf.constant([[0.0, 0.0], [1.0, 1.0]])),
            (tf.constant([[1.0, -1.0], [-1.0, 1.0]]), tf.constant([[1.0, 0.0], [0.0, 1.0]])),
        ],
    )
    def test_output_is_correct(self, non_negative_constraint, input, expected):
        output = non_negative_constraint(input)
        assert tf.math.reduce_all(output == expected)
